from typing import Optional, Any, Dict, Callable, get_type_hints
from pydantic import BaseModel, validate_call

from phi.utils.log import logger


class Message(BaseModel):
    """Pydantic model for holding LLM messages"""

    # The role of the messages author.
    # One of system, user, assistant, or function.
    role: str
    # The contents of the message. content is required for all messages,
    # and may be null for assistant messages with function calls.
    content: Optional[str] = None
    # The name of the author of this message. name is required if role is function,
    # and it should be the name of the function whose response is in the content.
    # May contain a-z, A-Z, 0-9, and underscores, with a maximum length of 64 characters.
    name: Optional[str] = None
    # The name and arguments of a function that should be called, as generated by the model.
    function_call: Optional[Any] = None
    # Metrics for the message, tokes + the time it took to generate the response.
    metrics: Dict[str, Any] = {}

    def to_dict(self) -> Dict[str, Any]:
        _dict = self.model_dump(exclude_none=True, exclude={"metrics"})
        # Manually add the content field if it is None
        if self.content is None:
            _dict["content"] = None
        return _dict

    def log(self, level: Optional[str] = None):
        """Log the message to the console

        @param level: The level to log the message at. One of debug, info, warning, or error.
            Defaults to debug.
        """
        _logger = logger.debug
        if level == "debug":
            _logger = logger.debug
        elif level == "info":
            _logger = logger.info
        elif level == "warning":
            _logger = logger.warning
        elif level == "error":
            _logger = logger.error

        if self.role == "function":
            _logger(f"{self.role.upper()}: {self.name}")
            _logger(f"{self.content}")
        else:
            _logger(f"{self.role.upper()}: {self.content or self.function_call}")


class References(BaseModel):
    """Pydantic model for holding LLM references"""

    # The question asked by the user.
    query: str
    # The references from the vector database.
    references: str
    # Performance in seconds.
    time: Optional[float] = None


class Function(BaseModel):
    """Pydantic model for holding LLM functions"""

    # The name of the function to be called.
    # Must be a-z, A-Z, 0-9, or contain underscores and dashes, with a maximum length of 64.
    name: str
    # A description of what the function does, used by the model to choose when and how to call the function.
    description: Optional[str] = None
    # The parameters the functions accepts, described as a JSON Schema object.
    # To describe a function that accepts no parameters, provide the value {"type": "object", "properties": {}}.
    parameters: Dict[str, Any] = {"type": "object", "properties": {}}
    entrypoint: Optional[Callable] = None

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump(exclude_none=True, exclude={"entrypoint"})

    @classmethod
    def from_callable(cls, c: Callable) -> "Function":
        from inspect import getdoc
        from phi.utils.json_schema import get_json_schema

        parameters = {"type": "object", "properties": {}}
        try:
            type_hints = get_type_hints(c)
            parameters = get_json_schema(type_hints)
            # logger.debug(f"Type hints for {c.__name__}: {type_hints}")
        except Exception as e:
            logger.warning(f"Could not parse args for {c.__name__}: {e}")

        return cls(
            name=c.__name__,
            description=getdoc(c),
            parameters=parameters,
            entrypoint=validate_call(c),
        )


class FunctionCall(BaseModel):
    """Pydantic model for holding LLM function calls"""

    # The function to be called.
    function: Function
    # The arguments to call the function with.
    arguments: Optional[Dict[str, Any]] = None
    # The result of the function call.
    result: Optional[Any] = None

    def get_call_str(self) -> str:
        """Returns a string representation of the function call."""
        if self.arguments is None:
            return f"{self.function.name}()"
        return f"{self.function.name}({', '.join([f'{k}={v}' for k, v in self.arguments.items()])})"

    def run(self) -> bool:
        """Runs the function call.

        @return: True if the function call was successful, False otherwise.
        """
        if self.function.entrypoint is None:
            return False

        logger.debug(f"Running: {self.get_call_str()}")

        # Call the function with no arguments if none are provided.
        if self.arguments is None:
            try:
                self.result = self.function.entrypoint()
                return True
            except Exception as e:
                logger.warning(f"Could not run function {self.get_call_str()}")
                logger.error(e)
                return False

        # Validate the arguments if provided.
        # try:
        #     from jsonschema import validate
        # except ImportError:
        #     raise ImportError("`jsonschema` is required for LLM functions, install using `pip install jsonschema`")
        try:
            self.result = self.function.entrypoint(**self.arguments)
            return True
        except Exception as e:
            logger.warning(f"Could not run function {self.get_call_str()}")
            logger.error(e)
            return False



import json
from typing import Optional, Dict, List, Any, Iterator

from rich import box
from rich.prompt import Prompt
from rich.live import Live
from rich.table import Table
from rich.markdown import Markdown

from phi.api.ai import conversation_chat
from phi.api.schemas.user import UserSchema
from phi.api.schemas.ai import ConversationType
from phi.cli.config import PhiCliConfig
from phi.cli.console import console
from phi.cli.settings import phi_cli_settings
from phi.llm.schemas import Function, Message, FunctionCall
from phi.llm.function.shell import ShellScriptsRegistry
from phi.llm.function.phi_commands import PhiCommandsRegistry
from phi.workspace.config import WorkspaceConfig
from phi.utils.log import logger
from phi.utils.functions import get_function_call
from phi.utils.timer import Timer
from phi.utils.json_io import write_json_file, read_json_file


class PhiAI:
    def __init__(
        self,
        new_conversation: bool = False,
        conversation_type: ConversationType = ConversationType.RAG,
        phi_config: Optional[PhiCliConfig] = None,
    ):
        logger.debug("--**-- Starting Phi AI --**--")

        _phi_config = phi_config or PhiCliConfig.from_saved_config()
        if _phi_config is None:
            raise ValueError("Phi config not found. Please run `phi init`")
        _user = _phi_config.user
        if _user is None:
            raise ValueError("User not found. Please run `phi auth`")
        _active_workspace = _phi_config.get_active_ws_config()

        self.conversation_db: Optional[List[Dict[str, Any]]] = None
        self.functions: Dict[str, Function] = {**ShellScriptsRegistry().functions, **PhiCommandsRegistry().functions}

        _conversation_id = None
        _conversation_history = None
        if not new_conversation:
            latest_conversation = self.get_latest_conversation()
            if latest_conversation is not None:
                # Check if the latest conversation is of the same type and user
                if latest_conversation.get("conversation_type") == conversation_type and latest_conversation.get(
                    "user"
                ) == _user.model_dump(include={"id_user", "email"}):
                    logger.debug("Found conversation for the same user and type")
                    # Check if the latest conversation is for the same workspace
                    if _active_workspace is not None and latest_conversation.get(
                        "workspace"
                    ) == _active_workspace.model_dump(include={"ws_dir_name"}):
                        logger.debug("Found conversation for the same workspace")
                        # Check if the latest conversation has the same functions
                        if latest_conversation.get("functions") == list(self.functions.keys()):
                            logger.debug("Found conversation with the same functions")
                            _conversation_id = latest_conversation.get("conversation_id")
                            _conversation_history = latest_conversation.get("conversation_history")

        if _conversation_id is None:
            from phi.api.ai import conversation_create, ConversationCreateResponse

            _conversation: Optional[ConversationCreateResponse] = conversation_create(
                user=_user, conversation_type=conversation_type, functions=self.functions
            )
            if _conversation is None:
                logger.error("Could not create conversation, please authenticate using `phi auth`")
                exit(0)

            _conversation_id = _conversation.id
            _conversation_history = _conversation.chat_history

        self.phi_config: PhiCliConfig = _phi_config
        self.user: UserSchema = _user
        self.active_workspace: Optional[WorkspaceConfig] = _active_workspace
        self.conversation_id: int = _conversation_id
        self.conversation_history: List[Dict[str, Any]] = _conversation_history or []
        self.conversation_type: ConversationType = conversation_type

        self.save_conversation()
        logger.debug(f"--**-- Conversation: {self.conversation_id} --**--")

    def start_conversation(self, stream: bool = False):
        conversation_active = True
        while conversation_active:
            username = self.user.username or "You"
            console.rule()
            user_message_str_valid = False
            while not user_message_str_valid:
                user_message_str = Prompt.ask(f"[bold] :sunglasses: {username} [/bold]", console=console)
                if (
                    user_message_str
                    is None
                    # or user_message_str == ""
                    # or user_message_str == "{}"
                    # or len(user_message_str) == 0
                ):
                    console.print("Please enter a valid message")
                    continue
                user_message_str_valid = True
            self.conversation_history.append({"role": "user", "content": user_message_str})

            # -*- Quit conversation
            if user_message_str in ("exit", "quit", "bye"):
                conversation_active = False

            # -*- Send message to Phi AI
            api_response: Optional[Iterator[str]] = conversation_chat(
                user=self.user,
                conversation_id=self.conversation_id,
                message=Message(role="user", content=user_message_str),
                conversation_type=self.conversation_type,
                functions=self.functions,
                stream=stream,
            )
            if api_response is None:
                logger.error("Could not reach Phi AI")
                conversation_active = False
            else:
                with Live(console=console) as live:
                    response_content = ""
                    if stream:
                        for _response in api_response:
                            if _response is None or _response == "" or _response == "{}":
                                continue
                            response_dict = json.loads(_response)
                            if "content" in response_dict and response_dict.get("content") is not None:
                                response_content += response_dict.get("content")
                                table = Table(show_header=False, box=box.ROUNDED)
                                table.add_row(Markdown(response_content))
                                live.update(table)
                            elif "function_call" in response_dict:
                                for function_response in self.run_function_stream(response_dict.get("function_call")):
                                    response_content += function_response
                                    table = Table(show_header=False, box=box.ROUNDED)
                                    table.add_row(Markdown(response_content))
                                    live.update(table)
                    else:
                        _response = next(api_response)
                        if _response is None or _response == "" or _response == "{}":
                            response_content = "Something went wrong, please try again."
                        else:
                            response_dict = json.loads(_response)
                            if "content" in response_dict and response_dict.get("content") is not None:
                                response_content = response_dict.get("content")
                            elif "function_call" in response_dict:
                                response_content = self.run_function(response_dict.get("function_call"))
                        table = Table(show_header=False, box=box.ROUNDED)
                        table.add_row(Markdown(response_content))
                        console.print(table)
                    self.conversation_history.append({"role": "assistant", "content": response_content})
            self.save_conversation()

    def run_function_stream(self, function_call: Dict[str, Any]) -> Iterator[str]:
        _function_name = function_call.get("name")
        _function_arguments_str = function_call.get("arguments")
        if _function_name is not None:
            function_call_obj: Optional[FunctionCall] = get_function_call(
                name=_function_name, arguments=_function_arguments_str, functions=self.functions
            )
            if function_call_obj is None:
                return "Something went wrong, please try again."

            # -*- Run function call
            yield f"Running: {function_call_obj.get_call_str()}\n\n"
            function_call_timer = Timer()
            function_call_timer.start()
            function_call_obj.run()
            function_call_timer.stop()
            function_call_message = Message(
                role="function",
                name=function_call_obj.function.name,
                content=function_call_obj.result,
                metrics={"time": function_call_timer.elapsed},
            )
            # -*- Send message to Phi AI
            api_response: Optional[Iterator[str]] = conversation_chat(
                user=self.user,
                conversation_id=self.conversation_id,
                message=function_call_message,
                conversation_type=self.conversation_type,
                functions=self.functions,
                stream=True,
            )
            if api_response is not None:
                for _response in api_response:
                    if _response is None or _response == "" or _response == "{}":
                        continue
                    response_dict = json.loads(_response)
                    if "content" in response_dict and response_dict.get("content") is not None:
                        yield response_dict.get("content")
                    elif "function_call" in response_dict:
                        yield from self.run_function_stream(response_dict.get("function_call"))
        else:
            yield "Could not run function, please try again."

    def run_function(self, function_call: Dict[str, Any]) -> str:
        _function_name = function_call.get("name")
        _function_arguments_str = function_call.get("arguments")
        if _function_name is not None:
            function_call_obj: Optional[FunctionCall] = get_function_call(
                name=_function_name, arguments=_function_arguments_str, functions=self.functions
            )
            if function_call_obj is None:
                return "Something went wrong, please try again."

            # -*- Run function call
            function_run_response = f"Running: {function_call_obj.get_call_str()}\n\n"
            function_call_timer = Timer()
            function_call_timer.start()
            function_call_obj.run()
            function_call_timer.stop()
            function_call_message = Message(
                role="function",
                name=function_call_obj.function.name,
                content=function_call_obj.result,
                metrics={"time": function_call_timer.elapsed},
            )
            # -*- Send message to Phi AI
            api_response: Optional[Iterator[str]] = conversation_chat(
                user=self.user,
                conversation_id=self.conversation_id,
                message=function_call_message,
                conversation_type=self.conversation_type,
                functions=self.functions,
                stream=False,
            )
            if api_response is not None:
                _response = next(api_response)
                if _response is None or _response == "" or _response == "{}":
                    function_run_response += "Something went wrong, please try again."
                else:
                    response_dict = json.loads(_response)
                    if "content" in response_dict and response_dict.get("content") is not None:
                        function_run_response += response_dict.get("content")
                    elif "function_call" in response_dict:
                        function_run_response += self.run_function(response_dict.get("function_call"))
            return function_run_response
        return "Something went wrong, please try again."

    def print_conversation_history(self):
        from rich import box
        from rich.table import Table
        from rich.markdown import Markdown

        table = Table(show_header=False, box=box.ROUNDED, show_lines=True)
        table.add_column("Role", justify="right", style="cyan")
        table.add_column("Message", justify="left", style="magenta")

        for message in self.conversation_history:
            if message["role"] == "system":
                continue
            elif message["role"] == "assistant":
                table.add_row("phi", Markdown(message["content"]))
            elif message["role"] == "user":
                username = self.user.username or "You"
                table.add_row(username, Markdown(message["content"]))
            else:
                table.add_row(message["role"], Markdown(message["content"]))
        console.print(table)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "conversation_id": self.conversation_id,
            "conversation_history": self.conversation_history,
            "conversation_type": self.conversation_type,
            "user": self.user.model_dump(include={"id_user", "email"}),
            "workspace": self.active_workspace.model_dump(include={"ws_dir_name"})
            if self.active_workspace is not None
            else None,
            "functions": list(self.functions.keys()),
        }

    def save_conversation(self):
        if self.conversation_id is None:
            return

        all_conversations = self.get_all_conversations()
        # Pop the latest conversation if it exists
        if len(all_conversations) > 0:
            all_conversations.pop(0)

        # Add the latest conversation
        all_conversations.insert(0, self.to_dict())
        write_json_file(file_path=phi_cli_settings.ai_conversations_path, data=all_conversations)

    def get_all_conversations(self) -> List[Dict[str, Any]]:
        try:
            if self.conversation_db is not None:
                return self.conversation_db
            stored_conversations = read_json_file(file_path=phi_cli_settings.ai_conversations_path)
            if stored_conversations is not None and isinstance(stored_conversations, list):
                self.conversation_db = stored_conversations
                return self.conversation_db
        except Exception as e:
            logger.warning(f"Could not read conversations in {phi_cli_settings.ai_conversations_path}: {e}")
        return []

    def get_latest_conversation(self) -> Optional[Dict[str, Any]]:
        conversations = self.get_all_conversations()
        if len(conversations) == 0:
            return None
        return conversations[0]

    # async def conversation(self, stream: bool = False):
    #     from rich import box
    #     from rich.prompt import Prompt
    #     from rich.live import Live
    #     from rich.table import Table
    #     from rich.markdown import Markdown
    #     from phi.api.ai import ai_ws_connect
    #
    #     logger.info("Starting conversation with Phi AI")
    #
    #     conversation_active = True
    #     username = self.user.username or "You"
    #     async with ai_ws_connect(
    #         user=self.user,
    #         conversation_id=self.conversation_id,
    #         conversation_type=self.conversation_type,
    #         stream=stream,
    #     ) as ai_ws:
    #         while conversation_active:
    #             console.rule()
    #             user_message = Prompt.ask(f"[bold] :sunglasses: {username} [/bold]", console=console)
    #             self.conversation_history.append({"role": "user", "content": user_message})
    #
    #             # -*- Quit conversation
    #             if user_message in ("exit", "quit", "bye"):
    #                 conversation_active = False
    #
    #             # -*- Send message to Phi AI
    #             await ai_ws.send(user_message)
    #             with Live(console=console) as live:
    #                 if stream:
    #                     chat_response = ""
    #                     ai_response_chunk = await ai_ws.recv()
    #                     while ai_response_chunk is not None and ai_response_chunk != "AI_RESPONSE_STOP_STREAM":
    #                         chat_response += ai_response_chunk
    #                         table = Table(show_header=False, box=box.ROUNDED)
    #                         table.add_row(Markdown(chat_response))
    #                         live.update(table)
    #                         ai_response_chunk = await ai_ws.recv()
    #                         if ai_response_chunk is None or ai_response_chunk == "AI_RESPONSE_STOP_STREAM":
    #                             break
    #                         if ai_response_chunk.startswith("{"):
    #                             await ai_ws.send("function_result:one")
    #                     self.conversation_history.append({"role": "assistant", "content": chat_response})
    #                 else:
    #                     ai_response = await ai_ws.recv()
    #                     logger.info(f"ai_response: {type(ai_response)} | {ai_response}")
    #
    #                     if ai_response is None:
    #                         logger.error("Could not reach Phi AI")
    #                         conversation_active = False
    #                     chat_response = ai_response
    #                     table = Table(show_header=False, box=box.ROUNDED)
    #                     table.add_row(Markdown(chat_response))
    #                     console.print(table)
    #                     self.conversation_history.append({"role": "assistant", "content": chat_response})
    #             self.save_conversation()



from typing import Dict, Optional

from kubernetes.client.models.v1_label_selector import V1LabelSelector
from pydantic import Field

from phi.k8s.resource.base import K8sObject


class LabelSelector(K8sObject):
    """
    A label selector is a label query over a set of resources.
    The result of matchLabels and matchExpressions are ANDed.
    An empty label selector matches all objects.
    A null label selector matches no objects.

    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#labelselector-v1-meta
    """

    resource_type: str = "LabelSelector"

    # matchLabels is a map of {key,value} pairs.
    match_labels: Optional[Dict[str, str]] = Field(None, alias="matchLabels")

    def get_k8s_object(self) -> V1LabelSelector:
        # Return a V1LabelSelector object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_label_selector.py
        _v1_label_selector = V1LabelSelector(
            match_labels=self.match_labels,
        )
        return _v1_label_selector



from typing import Optional, Dict, List, Tuple

from phi.embedder.base import Embedder


class OpenAIEmbedder(Embedder):
    model: str = "text-embedding-ada-002"
    dimensions: int = 1536

    def _response(self, text: str):
        try:
            from openai import Embedding  # noqa: F401
        except ImportError:
            raise ImportError("`openai` not installed")

        return Embedding.create(input=text, model=self.model)

    def get_embedding(self, text: str) -> List[float]:
        response = self._response(text=text)
        if "data" not in response:
            return []

        return response["data"][0]["embedding"]

    def get_embedding_and_usage(self, text: str) -> Tuple[List[float], Optional[Dict]]:
        response = self._response(text=text)
        if "data" not in response:
            return [], None

        embedding = response["data"][0]["embedding"]
        usage = response.get("usage", None)
        return embedding, usage



from pathlib import Path
from importlib import metadata

from pydantic import field_validator, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic_core.core_schema import FieldValidationInfo

PHI_CLI_DIR: Path = Path.home().resolve().joinpath(".phi")


class PhiCliSettings(BaseSettings):
    app_name: str = "phi"
    app_version: str = metadata.version("phidata")

    auth_token_path: Path = PHI_CLI_DIR.joinpath("token")
    config_file_path: Path = PHI_CLI_DIR.joinpath("config")
    credentials_path: Path = PHI_CLI_DIR.joinpath("credentials")
    ai_conversations_path: Path = PHI_CLI_DIR.joinpath("ai_conversations.json")
    auth_token_cookie: str = "__phi_session"
    auth_token_header: str = "X-PHIDATA-AUTH-TOKEN"

    api_runtime: str = "prd"
    api_enabled: bool = True
    api_url: str = Field("https://api.phidata.com", validate_default=True)
    signin_url: str = Field("https://phidata.com/signin", validate_default=True)

    model_config = SettingsConfigDict(env_prefix="PHI_")

    @field_validator("api_runtime", mode="before")
    def validate_runtime_env(cls, v):
        """Validate api_runtime."""

        valid_api_runtimes = ["dev", "stg", "prd"]
        if v not in valid_api_runtimes:
            raise ValueError(f"Invalid api_runtime: {v}")

        return v

    @field_validator("signin_url", mode="before")
    def update_signin_url(cls, v, info: FieldValidationInfo):
        api_runtime = info.data["api_runtime"]
        if api_runtime == "dev":
            return "http://localhost:3000/signin"
        elif api_runtime == "stg":
            return "https://stgphi.com/signin"
        else:
            return "https://phidata.com/signin"

    @field_validator("api_url", mode="before")
    def update_api_url(cls, v, info: FieldValidationInfo):
        api_runtime = info.data["api_runtime"]
        if api_runtime == "dev":
            return "http://localhost:7070"
        elif api_runtime == "stg":
            return "https://api.stgphi.com"
        else:
            return "https://api.phidata.com"


phi_cli_settings = PhiCliSettings()



from typing import Tuple, Optional


def parse_resource_filter(
    resource_filter: str,
) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str], Optional[str]]:
    target_env: Optional[str] = None
    target_infra: Optional[str] = None
    target_group: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    filters = resource_filter.split(":")
    num_filters = len(filters)
    if num_filters >= 1:
        if filters[0] != "":
            target_env = filters[0]
    if num_filters >= 2:
        if filters[1] != "":
            target_infra = filters[1]
    if num_filters >= 3:
        if filters[2] != "":
            target_group = filters[2]
    if num_filters >= 4:
        if filters[3] != "":
            target_name = filters[3]
    if num_filters >= 5:
        if filters[4] != "":
            target_type = filters[4]

    return target_env, target_infra, target_group, target_name, target_type


def parse_k8s_resource_filter(
    resource_filter: str,
) -> Tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
    target_env: Optional[str] = None
    target_group: Optional[str] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    filters = resource_filter.split(":")
    num_filters = len(filters)
    if num_filters >= 1:
        if filters[0] != "":
            target_env = filters[0]
    if num_filters >= 2:
        if filters[1] != "":
            target_group = filters[1]
    if num_filters >= 3:
        if filters[2] != "":
            target_name = filters[2]
    if num_filters >= 4:
        if filters[3] != "":
            target_type = filters[3]

    return target_env, target_group, target_name, target_type



from abc import ABC, abstractmethod
from typing import Optional, List

from phi.conversation.schemas import ConversationRow


class ConversationStorage(ABC):
    @abstractmethod
    def create(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def read(self, conversation_id: str) -> Optional[ConversationRow]:
        raise NotImplementedError

    @abstractmethod
    def get_all_conversation_ids(self, user_name: str) -> List[str]:
        raise NotImplementedError

    @abstractmethod
    def get_all_conversations(self, user_name: str) -> List[ConversationRow]:
        raise NotImplementedError

    @abstractmethod
    def upsert(self, conversation: ConversationRow) -> Optional[ConversationRow]:
        raise NotImplementedError

    @abstractmethod
    def end(self, conversation_id: str) -> Optional[ConversationRow]:
        raise NotImplementedError

    @abstractmethod
    def delete(self) -> None:
        raise NotImplementedError



from typing import Optional

from phi.app.db_app import DbApp
from phi.docker.app.base import DockerApp, ContainerContext  # noqa: F401


class Redis(DockerApp, DbApp):
    # -*- App Name
    name: str = "redis"

    # -*- Image Configuration
    image_name: str = "redis"
    image_tag: str = "7.2.0"

    # -*- App Ports
    # Open a container port if open_port=True
    open_port: bool = True
    port_number: int = 6379

    # -*- Redis Volume
    # Create a volume for redis storage
    create_volume: bool = True
    # Path to mount the volume inside the container
    volume_container_path: str = "/data"

    # -*- Redis Configuration
    # Provide REDIS_PASSWORD as redis_password or REDIS_PASSWORD in secrets_file
    redis_password: Optional[str] = None
    # Provide REDIS_SCHEMA as redis_schema or REDIS_SCHEMA in secrets_file
    redis_schema: Optional[str] = None
    redis_driver: str = "redis"
    logging_level: str = "debug"

    def get_db_password(self) -> Optional[str]:
        return self.db_password or self.get_secret_from_file("REDIS_PASSWORD")

    def get_db_database(self) -> Optional[str]:
        return self.redis_schema or self.get_secret_from_file("REDIS_SCHEMA")

    def get_db_driver(self) -> Optional[str]:
        return self.redis_driver

    def get_db_host(self) -> Optional[str]:
        return self.get_container_name()

    def get_db_port(self) -> Optional[int]:
        return self.container_port

    def get_db_connection(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_database()
        driver = self.get_db_driver()
        host = self.get_db_host()
        port = self.get_db_port()
        return f"{driver}://{password_str}{host}:{port}/{schema}"

    def get_db_connection_local(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_database()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{password_str}{host}:{port}/{schema}"



from typing import Optional, Dict
from pathlib import Path


def get_python_objects_from_module(module_path: Path) -> Dict:
    """Returns a dictionary of python objects from a module"""
    import importlib.util
    from importlib.machinery import ModuleSpec

    # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    # Create a ModuleSpec
    module_spec: Optional[ModuleSpec] = importlib.util.spec_from_file_location("module", module_path)
    # Using the ModuleSpec create a module
    if module_spec:
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)  # type: ignore
        return module.__dict__
    else:
        return {}



from typing import List, Optional, Iterator

from pydantic import BaseModel, ConfigDict

from phi.document import Document
from phi.document.reader.base import Reader
from phi.vectordb import VectorDb
from phi.utils.log import logger


class KnowledgeBase(BaseModel):
    """Base class for LLM knowledge base"""

    # Reader to read the documents
    reader: Optional[Reader] = None
    # Vector db to store the knowledge base
    vector_db: Optional[VectorDb] = None
    # Number of relevant documents to return on search
    num_documents: int = 5
    # Number of documents to optimize the vector db on
    optimize_on: Optional[int] = 1000

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def document_lists(self) -> Iterator[List[Document]]:
        """Iterator that yields lists of documents in the knowledge base
        Each object yielded by the iterator is a list of documents.
        """
        raise NotImplementedError

    def search(self, query: str, num_documents: Optional[int] = None) -> List[Document]:
        """Returns relevant documents matching the query"""

        if self.vector_db is None:
            logger.warning("No vector db provided")
            return []

        _num_documents = num_documents or self.num_documents
        logger.debug(f"Getting {_num_documents} relevant documents for query: {query}")
        return self.vector_db.search(query=query, limit=_num_documents)

    def load(self, recreate: bool = False) -> None:
        """Load the knowledge base to the vector db

        TODO: Use upsert instead of insert
        """

        if self.vector_db is None:
            logger.warning("No vector db provided")
            return

        if recreate:
            logger.debug("Deleting collection")
            self.vector_db.delete()

        logger.debug("Creating collection")
        self.vector_db.create()

        logger.info("Loading knowledge base")
        num_documents = 0

        for document_list in self.document_lists:
            # Filter out documents which already exist in the vector db
            if not recreate:
                document_list = [document for document in document_list if not self.vector_db.doc_exists(document)]

            self.vector_db.insert(documents=document_list)
            num_documents += len(document_list)
        logger.info(f"Loaded {num_documents} documents to knowledge base")

        if self.optimize_on is not None and num_documents > self.optimize_on:
            logger.debug("Optimizing Vector DB")
            self.vector_db.optimize()

    def exists(self) -> bool:
        """Returns True if the knowledge base exists"""
        if self.vector_db is None:
            logger.warning("No vector db provided")
            return False
        return self.vector_db.exists()



from collections import OrderedDict
from pathlib import Path
from typing import Dict, List, Optional

from phi.cli.console import print_heading, print_info
from phi.cli.settings import phi_cli_settings
from phi.api.schemas.user import UserSchema
from phi.api.schemas.workspace import WorkspaceSchema, WorkspaceDelete
from phi.utils.log import logger
from phi.utils.pickle import pickle_object_to_file, unpickle_object_from_file
from phi.workspace.config import WorkspaceConfig


class PhiCliConfig:
    """The PhiCliConfig class manages user data for the phi cli"""

    def __init__(self, user_schema: Optional[UserSchema] = None) -> None:
        # Current user, populated after authenticating with the api
        # To add a user, use the user setter
        self._user: Optional[UserSchema] = user_schema

        # Active ws dir - used as the default for `phi` commands
        # To add an active workspace, use the active_ws_dir setter
        self._active_ws_dir: Optional[str] = None

        # Mapping from ws_dir_name to ws_config
        self.ws_config_map: Dict[str, WorkspaceConfig] = OrderedDict()

        # Quick access from ws_root -> ws_config
        self.path_to_ws_config_map: Dict[Path, WorkspaceConfig] = OrderedDict()

        # Save the config to disk once initialized
        self.save_config()

    ######################################################
    ## User functions
    ######################################################

    @property
    def user(self) -> Optional[UserSchema]:
        return self._user

    @user.setter
    def user(self, user: Optional[UserSchema]) -> None:
        """Sets the user"""
        if user is not None:
            logger.debug(f"Setting user to: {user.email}")
            clear_user_cache = (
                self._user is not None  # previous user is not None
                and self._user.email != "anon"  # previous user is not anon
                and (user.email != self._user.email or user.id_user != self._user.id_user)  # new user is different
            )
            self._user = user
            if clear_user_cache:
                self.clear_user_cache()
            self.save_config()

    def clear_user_cache(self) -> None:
        """Clears the user cache"""
        logger.debug("Clearing user cache")
        self.ws_config_map.clear()
        self.path_to_ws_config_map.clear()
        self._active_ws_dir = None
        phi_cli_settings.ai_conversations_path.unlink(missing_ok=True)
        logger.info("Workspaces cleared, please setup again using `phi ws setup`")

    ######################################################
    ## Workspace functions
    ######################################################

    @property
    def active_ws_dir(self) -> Optional[str]:
        return self._active_ws_dir

    @active_ws_dir.setter
    def active_ws_dir(self, ws_dir_name: Optional[str]) -> None:
        if ws_dir_name is not None:
            logger.debug(f"Setting active workspace to: {ws_dir_name}")
            self._active_ws_dir = ws_dir_name
            self.save_config()

    @property
    def available_ws(self) -> List[WorkspaceConfig]:
        return list(self.ws_config_map.values())

    def _add_or_update_ws_config(
        self,
        ws_dir_name: str,
        ws_root_path: Optional[Path] = None,
        ws_schema: Optional[WorkspaceSchema] = None,
    ) -> Optional[WorkspaceConfig]:
        """The main function to create, update or refresh the WorkspaceConfig.

        Notes:
        1. ws_name is the only required argument.
        2. This function does not call self.save_config().
            Remember to save_config() after calling this function.
        """

        # Validate ws_name is not None
        if ws_dir_name is None or not isinstance(ws_dir_name, str):
            return None

        ######################################################
        # Create new ws_config for ws_name if one does not exist
        ######################################################
        if ws_dir_name not in self.ws_config_map:
            logger.debug(f"Creating workspace for directory: {ws_dir_name}")
            new_workspace_config = WorkspaceConfig(
                ws_dir_name=ws_dir_name,
                ws_schema=ws_schema,
                ws_root_path=ws_root_path,
            )
            self.ws_config_map[ws_dir_name] = new_workspace_config
            if ws_root_path is not None:
                self.path_to_ws_config_map[ws_root_path] = new_workspace_config
                logger.debug(f"Workspace dir: {ws_root_path}")
            logger.debug(f"Workspace created for directory: {ws_dir_name}")

            # Return the new_workspace_config
            return new_workspace_config

        ######################################################
        # Update ws_config
        ######################################################
        logger.debug(f"Updating workspace at directory: {ws_dir_name}")
        # By this point there should be a WorkspaceConfig object for this ws_name
        existing_ws_config: Optional[WorkspaceConfig] = self.ws_config_map.get(ws_dir_name, None)
        if existing_ws_config is None:
            logger.error(f"Could not find workspace at directory: {ws_dir_name}, please try again.")
            return None

        # Update ws_dir_name, ws_root_path, ws_schema if provided and do not match the existing values
        if ws_dir_name is not None and existing_ws_config.ws_dir_name != ws_dir_name:
            existing_ws_config.ws_dir_name = ws_dir_name
        if ws_root_path is not None and existing_ws_config.ws_root_path != ws_root_path:
            existing_ws_config.ws_root_path = ws_root_path
        if ws_schema is not None and existing_ws_config.ws_schema != ws_schema:
            existing_ws_config.ws_schema = ws_schema

        logger.debug(f"Workspace updated: {ws_dir_name}")

        # Return the updated_ws_config
        return existing_ws_config

    ######################################################
    # END
    ######################################################

    def add_new_ws_to_config(
        self,
        ws_dir_name: str,
        ws_root_path: Path,
    ) -> Optional[WorkspaceConfig]:
        """Adds a newly created workspace to the PhiCliConfig"""
        ws_config = self._add_or_update_ws_config(
            ws_dir_name=ws_dir_name,
            ws_root_path=ws_root_path,
        )
        self.save_config()
        return ws_config

    def update_ws_config(
        self,
        ws_dir_name: str,
        ws_schema: Optional[WorkspaceSchema] = None,
        ws_root_path: Optional[Path] = None,
        set_as_active: bool = False,
    ) -> Optional[WorkspaceConfig]:
        """Updates WorkspaceConfig and returns True if successful"""
        ws_config = self._add_or_update_ws_config(
            ws_dir_name=ws_dir_name,
            ws_schema=ws_schema,
            ws_root_path=ws_root_path,
        )
        if set_as_active:
            self.active_ws_dir = ws_dir_name
        self.save_config()
        return ws_config

    def delete_ws(self, ws_dir_name: str) -> None:
        """Handles Deleting a workspace from the PhiCliConfig and api"""

        print_heading(f"Deleting record for directory: {ws_dir_name}")
        print_info("-*- Note: this does not delete any files on disk, please delete them manually")

        ws_config: Optional[WorkspaceConfig] = self.ws_config_map.pop(ws_dir_name, None)
        if ws_config is None:
            logger.warning(f"No record of workspace at {ws_dir_name}")
            return

        if ws_config.ws_root_path is not None:
            self.path_to_ws_config_map.pop(ws_config.ws_root_path, None)

        # Check if we're deleting the active workspace, if yes, unset the active ws
        if (
            self._active_ws_dir is not None
            and ws_config.ws_dir_name is not None
            and self._active_ws_dir == ws_config.ws_dir_name
        ):
            print_info(f"Removing {ws_config.ws_dir_name} as the active workspace")
            self._active_ws_dir = None

        if self.user is not None and ws_config.ws_schema is not None:
            print_info(f"Deleting workspace {ws_config.ws_dir_name} from the server")

            from phi.api.workspace import delete_workspace_for_user

            delete_workspace_for_user(
                user=self.user,
                workspace=WorkspaceDelete(
                    id_workspace=ws_config.ws_schema.id_workspace, ws_name=ws_config.ws_schema.ws_name
                ),
            )

        self.save_config()

    ######################################################
    ## Get Workspace Data
    ######################################################

    def get_ws_config_by_dir_name(self, ws_dir_name: str) -> Optional[WorkspaceConfig]:
        return self.ws_config_map[ws_dir_name] if ws_dir_name in self.ws_config_map else None

    def get_ws_config_by_path(self, ws_root_path: Path) -> Optional[WorkspaceConfig]:
        return self.path_to_ws_config_map[ws_root_path] if ws_root_path in self.path_to_ws_config_map else None

    def get_ws_dir_name_by_path(self, ws_root_path: Path) -> Optional[str]:
        if ws_root_path in self.path_to_ws_config_map:
            return self.path_to_ws_config_map[ws_root_path].ws_dir_name
        return None

    def get_ws_schema_by_dir_name(self, ws_dir_name: str) -> Optional[WorkspaceSchema]:
        if ws_dir_name in self.ws_config_map:
            return self.ws_config_map[ws_dir_name].ws_schema
        return None

    def get_ws_schema_by_path(self, ws_root_path: Path) -> Optional[WorkspaceSchema]:
        if ws_root_path in self.path_to_ws_config_map:
            return self.path_to_ws_config_map[ws_root_path].ws_schema
        return None

    def get_ws_root_path_by_dir_name(self, ws_dir_name: str) -> Optional[Path]:
        if ws_dir_name in self.ws_config_map:
            return self.ws_config_map[ws_dir_name].ws_root_path
        return None

    def get_active_ws_config(self) -> Optional[WorkspaceConfig]:
        if self.active_ws_dir is not None and self.active_ws_dir in self.ws_config_map:
            return self.ws_config_map[self.active_ws_dir]
        return None

    ######################################################
    ## Save PhiCliConfig
    ######################################################

    def save_config(self):
        logger.debug(f"Saving config to {str(phi_cli_settings.config_file_path)}")
        pickle_object_to_file(self, phi_cli_settings.config_file_path)

    @classmethod
    def from_saved_config(cls):
        logger.debug(f"Reading PhiCliConfig from {phi_cli_settings.config_file_path}")
        try:
            return unpickle_object_from_file(file_path=phi_cli_settings.config_file_path, verify_class=cls)
        except Exception as e:
            logger.warning(e)
            logger.warning("If error persists, setup the workspace again using `phi ws setup`")

    ######################################################
    ## Print PhiCliConfig
    ######################################################

    def print_to_cli(self, show_all: bool = False):
        if self.user:
            print_heading(f"User: {self.user.email}\n")
        if self.active_ws_dir:
            print_heading(f"Active workspace directory: {self.active_ws_dir}\n")
        else:
            print_info("No active workspace found.")
            print_info("Please create using `phi ws create` or setup existing workspace using `phi ws setup`")

        if show_all and len(self.ws_config_map) > 0:
            print_heading("Available workspaces:\n")
            c = 1
            for k, v in self.ws_config_map.items():
                print_info(f"  {c}. Directory: {v.ws_dir_name}")
                print_info(f"     Path: {v.ws_root_path}")
                if v.ws_schema and v.ws_schema.ws_name:
                    print_info(f"     Name: {v.ws_schema.ws_name}")
                c += 1



import json
from uuid import uuid4
from datetime import datetime
from typing import List, Any, Optional, Dict, Iterator, Callable, cast, Union

from pydantic import BaseModel, ConfigDict, field_validator, model_validator, Field

from phi.conversation.memory import ConversationMemory
from phi.conversation.schemas import ConversationRow
from phi.conversation.storage import ConversationStorage
from phi.document import Document
from phi.knowledge.base import KnowledgeBase
from phi.llm.base import LLM
from phi.llm.openai import OpenAIChat
from phi.llm.schemas import Message, References
from phi.llm.function.registry import FunctionRegistry
from phi.utils.format_str import remove_indent
from phi.utils.log import logger, set_log_level_to_debug
from phi.utils.timer import Timer


class Conversation(BaseModel):
    # -*- LLM to use for this conversation
    llm: LLM = OpenAIChat()
    # -*- LLM Introduction
    # Add an introduction (from the LLM) to the chat history
    introduction: Optional[str] = None

    # -*- User settings
    # Name and type of user participating in this conversation.
    user_name: Optional[str] = None
    user_type: Optional[str] = None

    # -*- Conversation settings
    # Conversation UUID
    id: str = Field(default_factory=lambda: str(uuid4()))
    # Conversation name
    name: Optional[str] = None
    # True if this conversation is active i.e. not ended
    is_active: bool = True
    # Metadata associated with this conversation
    meta_data: Optional[Dict[str, Any]] = None
    # Extra data associated with this conversation
    extra_data: Optional[Dict[str, Any]] = None
    # The timestamp of when this conversation was created in the database
    created_at: Optional[datetime] = None
    # The timestamp of when this conversation was last updated in the database
    updated_at: Optional[datetime] = None

    # Monitor conversations on phidata.com
    monitoring: bool = False

    # -*- Conversation Memory
    memory: ConversationMemory = ConversationMemory()
    # Add chat history to the prompt sent to the LLM.
    # If True, a formatted chat history is added to the default user_prompt.
    add_chat_history_to_prompt: bool = False
    # Add chat history to the messages sent to the LLM.
    # If True, the chat history is added to the messages sent to the LLM.
    add_chat_history_to_messages: bool = False
    # Number of previous messages to add to prompt or messages sent to the LLM.
    num_history_messages: int = 8

    # -*- Conversation Knowledge Base
    knowledge_base: Optional[KnowledgeBase] = None
    # Add references from the knowledge base to the prompt sent to the LLM.
    add_references_to_prompt: bool = False

    # -*- Conversation Storage
    storage: Optional[ConversationStorage] = None
    # Create table if it doesn't exist
    create_storage: bool = True
    # ConversationRow from the database: DO NOT SET THIS MANUALLY
    database_row: Optional[ConversationRow] = None

    # -*- Enable Function Calls
    # Makes the conversation autonomous.
    function_calls: bool = False
    # Add a list of default functions to the LLM
    default_functions: bool = True
    # Show function calls in LLM messages.
    show_function_calls: bool = False
    # A list of functions to add to the LLM.
    functions: Optional[List[Callable]] = None
    # A list of function registries to add to the LLM.
    function_registries: Optional[List[FunctionRegistry]] = None

    #
    # -*- Prompt Settings
    #
    # -*- System prompt: provide the system prompt as a string or using a function
    system_prompt: Optional[str] = None
    # Function to build the system prompt.
    # This function is provided the conversation as an argument
    #   and should return the system_prompt as a string.
    # Signature:
    # def system_prompt_function(conversation: Conversation) -> str:
    #    ...
    system_prompt_function: Optional[Callable[..., Optional[str]]] = None
    # -*- User prompt: provide the user prompt as a string or using a function
    user_prompt: Optional[str] = None
    # Function to build the user prompt.
    # This function is provided the conversation and the user message as arguments
    #   and should return the user_prompt as a string.
    # If add_references_to_prompt is True, then references are also provided as an argument.
    # If add_chat_history_to_prompt is True, then chat_history is also provided as an argument.
    # Signature:
    # def custom_user_prompt_function(
    #     conversation: Conversation,
    #     message: str,
    #     references: Optional[str] = None,
    #     chat_history: Optional[str] = None,
    # ) -> str:
    #     ...
    user_prompt_function: Optional[Callable[..., str]] = None

    # -*- Functions to customize the user_prompt
    # Function to build references for the default user_prompt
    # This function, if provided, is called when add_references_to_prompt is True
    # Signature:
    # def references(conversation: Conversation, query: str) -> Optional[str]:
    #     ...
    references_function: Optional[Callable[..., Optional[str]]] = None
    # Function to build the chat_history for the default user prompt
    # This function, if provided, is called when add_chat_history_to_prompt is True
    # Signature:
    # def chat_history(conversation: Conversation) -> str:
    #     ...
    chat_history_function: Optional[Callable[..., Optional[str]]] = None

    # If True, show debug logs
    debug_mode: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v:
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    @model_validator(mode="after")
    def add_functions_to_llm(self) -> "Conversation":
        if self.function_calls:
            if self.default_functions:
                default_func_list: List[Callable] = [
                    self.get_last_n_chats,
                    self.search_knowledge_base,
                ]
                for func in default_func_list:
                    self.llm.add_function(func)

            # Add functions from self.functions
            if self.functions is not None:
                for func in self.functions:
                    self.llm.add_function(func)

            # Add functions from registries
            if self.function_registries is not None:
                for registry in self.function_registries:
                    self.llm.add_function_registry(registry)

            # Set function call to auto if it is not set
            if self.llm.function_call is None:
                self.llm.function_call = "auto"

            # Set show_function_calls if it is not set on the llm
            if self.llm.show_function_calls is None:
                self.llm.show_function_calls = self.show_function_calls
        return self

    def to_database_row(self) -> ConversationRow:
        """Create a ConversationRow for the current conversation (to save to the database)"""

        return ConversationRow(
            id=self.id,
            name=self.name,
            user_name=self.user_name,
            user_type=self.user_type,
            is_active=self.is_active,
            llm=self.llm.to_dict(),
            memory=self.memory.to_dict(),
            meta_data=self.meta_data,
            extra_data=self.extra_data,
            created_at=self.created_at,
            updated_at=self.updated_at,
        )

    def from_database_row(self, row: ConversationRow):
        """Load the existing conversation from a ConversationRow (from the database)"""

        # Values that are overwritten from the database if they are not set in the conversation
        if self.id is None and row.id is not None:
            self.id = row.id
        if self.name is None and row.name is not None:
            self.name = row.name
        if self.user_name is None and row.user_name is not None:
            self.user_name = row.user_name
        if self.user_type is None and row.user_type is not None:
            self.user_type = row.user_type
        if self.is_active is None and row.is_active is not None:
            self.is_active = row.is_active

        # Update llm data from the ConversationRow
        if row.llm is not None:
            # Update llm metrics from the database
            llm_metrics_from_db = row.llm.get("metrics")
            if llm_metrics_from_db is not None and isinstance(llm_metrics_from_db, dict):
                try:
                    self.llm.metrics = llm_metrics_from_db
                except Exception as e:
                    logger.warning(f"Failed to load llm metrics: {e}")

            # # Update llm functions from the database
            # llm_functions_from_db = row.llm.get("functions")
            # if llm_functions_from_db is not None and isinstance(llm_functions_from_db, dict):
            #     try:
            #         for k, v in llm_functions_from_db.items():
            #             _llm_function = Function(**v)
            #             self.llm.add_function_schema(func=_llm_function, if_not_exists=True)
            #     except Exception as e:
            #         logger.error(f"Failed to load llm functions: {e}")

        # Update conversation memory from the ConversationRow
        if row.memory is not None:
            try:
                self.memory = self.memory.__class__.model_validate(row.memory)
            except Exception as e:
                logger.warning(f"Failed to load conversation memory: {e}")

        # Update meta_data from the database
        if row.meta_data is not None:
            # If meta_data is set in the conversation,
            # merge it with the database meta_data. The conversation meta_data takes precedence
            if self.meta_data is not None and row.meta_data is not None:
                self.meta_data = {**row.meta_data, **self.meta_data}
            # If meta_data is not set in the conversation, use the database meta_data
            if self.meta_data is None and row.meta_data is not None:
                self.meta_data = row.meta_data

        # Update extra_data from the database
        if row.extra_data is not None:
            # If extra_data is set in the conversation,
            # merge it with the database extra_data. The conversation extra_data takes precedence
            if self.extra_data is not None and row.extra_data is not None:
                self.extra_data = {**row.extra_data, **self.extra_data}
            # If extra_data is not set in the conversation, use the database extra_data
            if self.extra_data is None and row.extra_data is not None:
                self.extra_data = row.extra_data

        # Update the timestamp of when this conversation was created in the database
        if row.created_at is not None:
            self.created_at = row.created_at

        # Update the timestamp of when this conversation was last updated in the database
        if row.updated_at is not None:
            self.updated_at = row.updated_at

    def read_from_storage(self) -> Optional[ConversationRow]:
        """Load the conversation from storage"""

        if self.storage is not None and self.id is not None:
            self.database_row = self.storage.read(conversation_id=self.id)
            if self.database_row is not None:
                logger.debug(f"-*- Loading conversation: {self.database_row.id}")
                self.from_database_row(row=self.database_row)
                logger.debug(f"-*- Loaded conversation: {self.id}")
        return self.database_row

    def write_to_storage(self) -> Optional[ConversationRow]:
        """Save the conversation to the storage"""

        if self.storage is not None:
            if self.create_storage:
                self.storage.create()
            self.database_row = self.storage.upsert(conversation=self.to_database_row())
        return self.database_row

    def add_introduction(self, introduction: str) -> None:
        """Add the introduction to the chat history"""

        if introduction is not None:
            if len(self.memory.chat_history) == 0:
                self.memory.add_chat_message(Message(role="assistant", content=introduction))

    def start(self) -> Optional[str]:
        """Start the conversation and return the conversation ID. This function:
        - Creates a new conversation in the storage if it does not exist
        - Load the conversation from the storage if it exists
        """

        # If a database_row exists, return the conversation_id
        if self.database_row is not None:
            return self.database_row.id

        # Create a new conversation or load an existing conversation
        if self.storage is not None:
            # Load existing conversation if it exists
            logger.debug(f"Reading conversation: {self.id}")
            self.read_from_storage()

            # Create a new conversation
            if self.database_row is None:
                logger.debug("-*- Creating new conversation")
                if self.introduction:
                    self.add_introduction(self.introduction)
                self.database_row = self.write_to_storage()
                if self.database_row is None:
                    raise Exception("Failed to create new conversation in storage")
                logger.debug(f"-*- Created conversation: {self.database_row.id}")
                self.from_database_row(row=self.database_row)
                self._api_upsert_conversation()
        return self.id

    def end(self) -> None:
        """End the conversation"""
        if self.storage is not None and self.id is not None:
            self.storage.end(conversation_id=self.id)
        self.is_active = False

    def get_system_prompt(self) -> str:
        """Return the system prompt for the conversation"""

        if self.system_prompt is not None:
            return "\n".join([line.strip() for line in self.system_prompt.split("\n")])

        if self.system_prompt_function is not None:
            system_prompt_kwargs = {"conversation": self}
            _system_prompt_from_function = remove_indent(self.system_prompt_function(**system_prompt_kwargs))
            if _system_prompt_from_function is not None:
                return _system_prompt_from_function
            else:
                raise Exception("system_prompt_function returned None")

        _system_prompt = "You are a chatbot "

        if self.user_type:
            _system_prompt += f"that is designed to help a '{self.user_type}' with their work.\n"
        else:
            _system_prompt += "that is designed to help a user with their work.\n"

        if self.knowledge_base is not None:
            _system_prompt += "You have access to a knowledge base that you can search for information.\n"
            _system_prompt += (
                "If you cannot find the answer in the knowledge base, "
                "say that you cannot find the answer in the knowledge base.\n"
            )

        if self.function_calls:
            _system_prompt += "You have access to functions that you can run to help you respond to the user.\n"

        _system_prompt += """Remember the following guidelines:
        - If you don't know the answer, say 'I don't know'.
        - Use bullet points where possible.
        - User markdown to format your answers.
        """

        # Return the system prompt after removing newlines and indenting
        _system_prompt = cast(str, remove_indent(_system_prompt))
        return _system_prompt

    def get_references_from_knowledge_base(self, query: str, num_documents: Optional[int] = None) -> Optional[str]:
        """Return a list of references from the knowledge base"""

        if self.references_function is not None:
            reference_kwargs = {"conversation": self, "query": query}
            return remove_indent(self.references_function(**reference_kwargs))

        if self.knowledge_base is None:
            return None

        relevant_docs: List[Document] = self.knowledge_base.search(query=query, num_documents=num_documents)
        return json.dumps([doc.to_dict() for doc in relevant_docs])

    def get_formatted_chat_history(self) -> Optional[str]:
        """Returns a formatted chat history to use in the user prompt"""

        if self.chat_history_function is not None:
            chat_history_kwargs = {"conversation": self}
            return remove_indent(self.chat_history_function(**chat_history_kwargs))

        formatted_history = self.memory.get_formatted_chat_history(num_messages=self.num_history_messages)
        if formatted_history == "":
            return None
        return remove_indent(formatted_history)

    def get_user_prompt(
        self, message: str, references: Optional[str] = None, chat_history: Optional[str] = None
    ) -> str:
        """Build the user prompt given a message, references and chat_history"""

        if self.user_prompt_function is not None:
            user_prompt_kwargs = {
                "conversation": self,
                "message": message,
                "references": references,
                "chat_history": chat_history,
            }
            _user_prompt_from_function = remove_indent(self.user_prompt_function(**user_prompt_kwargs))
            if _user_prompt_from_function is not None:
                return _user_prompt_from_function
            else:
                raise Exception("user_prompt_function returned None")

        _user_prompt = "Your task is to respond to the following message"
        if self.user_type:
            _user_prompt += f" from a '{self.user_type}'"
        _user_prompt += " in the best way possible.\n"

        # Add references to prompt
        if references:
            _user_prompt += f"""\n
                You can use the following information from the knowledge base if it helps respond to the message.
                START OF KNOWLEDGE BASE INFORMATION
                ```
                {references}
                ```
                END OF KNOWLEDGE BASE INFORMATION
                """

        # Add chat_history to prompt
        if chat_history:
            _user_prompt += f"""\n
                You can use the following chat history to reference past messages.
                START OF CHAT HISTORY
                ```
                {chat_history}
                ```
                END OF CHAT HISTORY
                """

        # Remind the LLM of its task
        if references or chat_history:
            _user_prompt += "\nRemember, your task is to respond to the following message in the best way possible."

        _user_prompt += f"\nUSER: {message}"
        _user_prompt += "\nASSISTANT: "

        # Return the user prompt after removing newlines and indenting
        _user_prompt = cast(str, remove_indent(_user_prompt))
        return _user_prompt

    def _chat(self, message: str, stream: bool = True) -> Iterator[str]:
        logger.debug("*********** Conversation Chat Start ***********")
        # Load the conversation from the database if available
        self.read_from_storage()

        # -*- Build the system prompt
        system_prompt = self.get_system_prompt()

        # -*- References to add to the user_prompt and send to the api for monitoring
        references: Optional[References] = None

        # -*- Get references to add to the user_prompt
        user_prompt_references = None
        if self.add_references_to_prompt:
            reference_timer = Timer()
            reference_timer.start()
            user_prompt_references = self.get_references_from_knowledge_base(query=message)
            reference_timer.stop()
            references = References(
                query=message, references=user_prompt_references, time=round(reference_timer.elapsed, 4)
            )
            logger.debug(f"Time to get references: {reference_timer.elapsed:.4f}s")

        # -*- Get chat history to add to the user prompt
        user_prompt_chat_history = None
        if self.add_chat_history_to_prompt:
            user_prompt_chat_history = self.get_formatted_chat_history()

        # -*- Build the user prompt
        user_prompt = self.get_user_prompt(
            message=message, references=user_prompt_references, chat_history=user_prompt_chat_history
        )

        # -*- Build the messages to send to the LLM
        # Create system message
        system_prompt_message = Message(role="system", content=system_prompt)
        # Create user message
        user_prompt_message = Message(role="user", content=user_prompt)

        # Create message list
        messages: List[Message] = [system_prompt_message]
        if self.add_chat_history_to_messages:
            messages += self.memory.get_last_n_messages(last_n=self.num_history_messages)
        messages += [user_prompt_message]

        # -*- Generate response (includes running function calls)
        llm_response = ""
        if stream:
            for response_chunk in self.llm.parsed_response_stream(messages=messages):
                llm_response += response_chunk
                yield response_chunk
        else:
            llm_response = self.llm.parsed_response(messages=messages)

        # -*- Add messages to the memory
        # Add the system prompt to the memory - added only if this is the first message to the LLM
        self.memory.add_system_prompt(message=system_prompt_message)

        # Add user message to the memory - this is added to the chat_history
        self.memory.add_user_message(message=Message(role="user", content=message))

        # Add user prompt to the memory - this is added to the llm_messages
        self.memory.add_llm_message(message=user_prompt_message)

        # Add references to the memory
        if references:
            self.memory.add_references(references=references)

        # Add llm response to the memory - this is added to the chat_history and llm_messages
        self.memory.add_llm_response(message=Message(role="assistant", content=llm_response))

        # -*- Save conversation to storage
        self.write_to_storage()

        # -*- Send conversation event for monitoring
        event_data = {
            "user_message": message,
            "llm_response": llm_response,
            "messages": [m.model_dump(exclude_none=True) for m in messages],
            "references": references.model_dump(exclude_none=True) if references else None,
            "metrics": self.llm.metrics,
        }
        self._api_send_conversation_event(event_type="chat", event_data=event_data)

        # -*- Yield final response if not streaming
        if not stream:
            yield llm_response
        logger.debug("*********** Conversation Chat End ***********")

    def chat(self, message: str, stream: bool = True) -> Union[Iterator[str], str]:
        resp = self._chat(message=message, stream=stream)
        if stream:
            return resp
        else:
            return next(resp)

    def _chat_raw(
        self, messages: List[Message], user_message: Optional[str] = None, stream: bool = True
    ) -> Iterator[Dict]:
        logger.debug("*********** Conversation Raw Chat Start ***********")
        # Load the conversation from the database if available
        self.read_from_storage()

        # -*- Add user message to the memory - this is added to the chat_history
        if user_message:
            self.memory.add_user_message(Message(role="user", content=user_message))

        # -*- Add prompts to the memory - these are added to the llm_messages
        self.memory.add_llm_messages(messages=messages)

        # -*- Generate response
        batch_llm_response_message = {}
        if stream:
            for response_delta in self.llm.response_delta(messages=messages):
                yield response_delta
        else:
            batch_llm_response_message = self.llm.response_message(messages=messages)

        # Add llm response to the memory - this is added to the chat_history and llm_messages
        # LLM Response is the last message in the messages list
        llm_response_message = messages[-1]
        try:
            self.memory.add_llm_response(llm_response_message)
        except Exception as e:
            logger.warning(f"Failed to add llm response to memory: {e}")

        # -*- Save conversation to storage
        self.write_to_storage()

        # -*- Send conversation event for monitoring
        event_data = {
            "user_message": user_message,
            "llm_response": llm_response_message,
            "messages": [m.model_dump(exclude_none=True) for m in messages],
            "metrics": self.llm.metrics,
        }
        self._api_send_conversation_event(event_type="chat_raw", event_data=event_data)

        # -*- Yield final response if not streaming
        if not stream:
            yield batch_llm_response_message
        logger.debug("*********** Conversation Raw Chat End ***********")

    def chat_raw(
        self, messages: List[Message], user_message: Optional[str] = None, stream: bool = True
    ) -> Union[Iterator[Dict], Dict]:
        resp = self._chat_raw(messages=messages, user_message=user_message, stream=stream)
        if stream:
            return resp
        else:
            return next(resp)

    def rename(self, name: str) -> None:
        """Rename the conversation"""
        self.read_from_storage()
        self.name = name

        # -*- Save conversation to storage
        self.write_to_storage()

        # -*- Update conversation
        self._api_upsert_conversation()

    def generate_name(self) -> str:
        """Generate a name for the conversation using chat history"""
        _conv = (
            "Please provide a suitable name for the following conversation in maximum 5 words.\n"
            "Remember, do not exceed 5 words.\n\n"
        )
        for message in self.memory.chat_history[1:6]:
            _conv += f"{message.role.upper()}: {message.content}\n"

        _conv += "\n\nConversation Name:"

        system_message = Message(
            role="system",
            content="Please provide a suitable name for the following conversation in maximum 5 words.",
        )
        user_message = Message(role="user", content=_conv)
        generate_name_messages = [system_message, user_message]
        generated_name = self.llm.parsed_response(messages=generate_name_messages)
        if len(generated_name.split()) > 15:
            logger.error("Generated name is too long. Trying again.")
            return self.generate_name()
        return generated_name.replace('"', "").strip()

    def auto_rename(self) -> None:
        """Automatically rename the conversation"""
        self.read_from_storage()
        generated_name = self.generate_name()
        logger.debug(f"Generated name: {generated_name}")
        self.name = generated_name

        # -*- Save conversation to storage
        self.write_to_storage()

        # -*- Update conversation
        self._api_upsert_conversation()

    ###########################################################################
    # Api functions
    ###########################################################################

    def _api_upsert_conversation(self):
        if not self.monitoring:
            return

        from phi.api.conversation import upsert_conversation, ConversationUpdate

        logger.debug("Sending conversation event")
        try:
            database_row: ConversationRow = self.database_row or self.to_database_row()
            upsert_conversation(
                conversation=ConversationUpdate(
                    conversation_key=database_row.conversation_key(),
                    conversation_data=database_row.conversation_data(),
                ),
            )
        except Exception as e:
            logger.debug(f"Could not log conversation event: {e}")

    def _api_send_conversation_event(
        self, event_type: str = "chat", event_data: Optional[Dict[str, Any]] = None
    ) -> None:
        if not self.monitoring:
            return

        from phi.api.conversation import log_conversation_event, ConversationEventCreate

        logger.debug("Sending conversation event")
        try:
            database_row: ConversationRow = self.database_row or self.to_database_row()
            log_conversation_event(
                conversation=ConversationEventCreate(
                    conversation_key=database_row.conversation_key(),
                    conversation_data=database_row.conversation_data(),
                    event_type=event_type,
                    event_data=event_data,
                ),
            )
        except Exception as e:
            logger.debug(f"Could not log conversation event: {e}")

    ###########################################################################
    # LLM functions
    ###########################################################################

    def get_last_n_chats(self, num_chats: Optional[int] = None) -> str:
        """Returns the last n chats between the user and assistant.
        Example:
            - To get the last chat, use num_chats=1.
            - To get the last 5 chats, use num_chats=5.
            - To get all chats, use num_chats=None.
            - To get the first chat, use num_chats=None and pick the first message.
        :param num_chats: The number of chats to return.
            Each chat contains 2 messages. One from the user and one from the assistant.
        :return: A list of dictionaries representing the chat history.
        """
        history: List[Dict[str, Any]] = []
        all_chats = self.memory.get_chats()
        if len(all_chats) == 0:
            return ""

        chats_added = 0
        for chat in all_chats[::-1]:
            history.insert(0, chat[1].to_dict())
            history.insert(0, chat[0].to_dict())
            chats_added += 1
            if num_chats is not None and chats_added >= num_chats:
                break
        return json.dumps(history)

    def search_knowledge_base(self, query: str) -> Optional[str]:
        """Search the knowledge base for information about a users query.

        :param query: The query to search for.
        :return: A string containing the response from the knowledge base.
        """
        return self.get_references_from_knowledge_base(query=query)

    ###########################################################################
    # Print Response
    ###########################################################################

    def print_response(self, message: str, stream: bool = True) -> None:
        from phi.cli.console import console
        from rich.live import Live
        from rich.table import Table
        from rich.box import ROUNDED
        from rich.markdown import Markdown

        if stream:
            response = ""
            with Live() as live_log:
                for resp in self.chat(message, stream=True):
                    response += resp

                    table = Table(box=ROUNDED, border_style="blue")
                    table.add_column("Message")
                    table.add_column(message)
                    md_response = Markdown(response)
                    table.add_row("Response", md_response)
                    live_log.update(table)
        else:
            response = self.chat(message, stream=False)  # type: ignore
            md_response = Markdown(response)
            table = Table(box=ROUNDED, border_style="blue")
            table.add_column("Message")
            table.add_column(message)
            table.add_row("Response", md_response)
            console.print(table)



from datetime import datetime
from typing import Optional, Any, Dict
from pydantic import BaseModel, ConfigDict


class ConversationRow(BaseModel):
    """Interface between Conversation class and the database"""

    # Conversation UUID
    id: str
    # Conversation name
    name: Optional[str] = None
    # User settings
    # Name and type of user participating in this conversation.
    user_name: Optional[str] = None
    user_type: Optional[str] = None
    # True if this conversation is active i.e. not ended
    is_active: Optional[bool] = None
    # LLM data (name, model, etc.)
    llm: Optional[Dict[str, Any]] = None
    # Conversation Memory
    memory: Optional[Dict[str, Any]] = None
    # Metadata associated with this conversation
    meta_data: Optional[Dict[str, Any]] = None
    # Extra data associated with this conversation
    extra_data: Optional[Dict[str, Any]] = None
    # The timestamp of when this conversation was created
    created_at: Optional[datetime] = None
    # The timestamp of when this conversation was last updated
    updated_at: Optional[datetime] = None

    model_config = ConfigDict(from_attributes=True)

    def serializable_dict(self):
        _dict = self.model_dump(exclude={"created_at", "updated_at"})
        _dict["created_at"] = self.created_at.isoformat() if self.created_at else None
        _dict["updated_at"] = self.updated_at.isoformat() if self.updated_at else None
        return _dict

    def conversation_data(self) -> Dict[str, Any]:
        """Returns the conversation data as a dictionary."""
        _dict = self.model_dump(exclude={"memory", "created_at", "updated_at"})
        _dict["created_at"] = self.created_at.isoformat() if self.created_at else None
        _dict["updated_at"] = self.updated_at.isoformat() if self.updated_at else None
        return _dict

    def conversation_key(self) -> str:
        """Returns the conversation key."""
        keys = []
        if self.user_name:
            keys.append(self.user_name)
        if self.user_type:
            keys.append(self.user_type)
        if self.id:
            keys.append(str(self.id))
        return "::".join(keys)



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from pydantic import Field, field_serializer

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_service import V1Service
from kubernetes.client.models.v1_service_list import V1ServiceList
from kubernetes.client.models.v1_service_port import V1ServicePort
from kubernetes.client.models.v1_service_spec import V1ServiceSpec
from kubernetes.client.models.v1_service_status import V1ServiceStatus
from kubernetes.client.models.v1_status import V1Status

from phi.k8s.api_client import K8sApiClient
from phi.k8s.resource.base import K8sResource, K8sObject
from phi.k8s.enums.protocol import Protocol
from phi.k8s.enums.service_type import ServiceType
from phi.utils.log import logger


class ServicePort(K8sObject):
    """
    Reference:
    - Docs: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#serviceport-v1-core
    - Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service_port.py
    """

    resource_type: str = "ServicePort"

    # The name of this port within the service. This must be a DNS_LABEL.
    # All ports within a ServiceSpec must have unique names.
    # When considering the endpoints for a Service, this must match the 'name' field in the EndpointPort.
    # Optional if only one ServicePort is defined on this service.
    name: Optional[str] = None
    # The port on each node on which this service is exposed when type is NodePort or LoadBalancer.
    # Usually assigned by the system. If a value is specified, in-range, and not in use it will be used,
    # otherwise the operation will fail.
    # If not specified, a port will be allocated if this Service requires one.
    # If this field is specified when creating a Service which does not need it, creation will fail.
    # More info: https://kubernetes.io/docs/concepts/services-networking/service/#type-nodeport
    node_port: Optional[int] = Field(None, alias="nodePort")
    # The port that will be exposed by this service.
    port: int
    # The IP protocol for this port.
    # Supports "TCP", "UDP", and "SCTP". Default is TCP.
    protocol: Optional[Protocol] = None
    # Number or name of the port to access on the pods targeted by the service.
    # Number must be in the range 1 to 65535. Name must be an IANA_SVC_NAME.
    # If this is a string, it will be looked up as a named port in the target Pod's container ports.
    # If this is not specified, the value of the 'port' field is used (an identity map).
    # This field is ignored for services with clusterIP=None, and should be omitted or set equal to the 'port' field.
    # More info: https://kubernetes.io/docs/concepts/services-networking/service/#defining-a-service
    target_port: Optional[Union[str, int]] = Field(None, alias="targetPort")
    # The application protocol for this port. This field follows standard Kubernetes label syntax.
    app_protocol: Optional[str] = Field(None, alias="appProtocol")

    @field_serializer("protocol")
    def get_protocol_value(self, v) -> Optional[str]:
        return v.value if v else None

    def get_k8s_object(self) -> V1ServicePort:
        # logger.info(f"Building {self.get_resource_type()} : {self.get_resource_name()}")

        target_port_int: Optional[int] = None
        if isinstance(self.target_port, int):
            target_port_int = self.target_port
        elif isinstance(self.target_port, str):
            try:
                target_port_int = int(self.target_port)
            except ValueError:
                pass

        target_port = target_port_int or self.target_port
        # logger.info(f"target_port         : {type(self.target_port)} | {self.target_port}")
        # logger.info(f"target_port updated : {type(target_port)} | {target_port}")

        # Return a V1ServicePort object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service_port.py
        _v1_service_port = V1ServicePort(
            name=self.name,
            node_port=self.node_port,
            port=self.port,
            protocol=self.protocol.value if self.protocol else None,
            target_port=target_port,
            app_protocol=self.app_protocol,
        )
        return _v1_service_port


class ServiceSpec(K8sObject):
    """
    Reference:
    - https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#servicespec-v1-core
    """

    resource_type: str = "ServiceSpec"

    # type determines how the Service is exposed.
    # Defaults to ClusterIP. Valid options are ExternalName, ClusterIP, NodePort, and LoadBalancer.
    # "ClusterIP" allocates a cluster-internal IP address for load-balancing to endpoints.
    #   Endpoints are determined by the selector or if that is not specified,
    #   by manual construction of an Endpoints object or EndpointSlice objects.
    #   If clusterIP is "None", no virtual IP is allocated and the endpoints
    #   are published as a set of endpoints rather than a virtual IP.
    # "NodePort" builds on ClusterIP and allocates a port on every node which
    #   routes to the same endpoints as the clusterIP.
    # "LoadBalancer" builds on NodePort and creates an external load-balancer (if supported in the current cloud)
    #   which routes to the same endpoints as the clusterIP.
    # "ExternalName" aliases this service to the specified externalName.
    # Several other fields do not apply to ExternalName services.
    # More info: https://kubernetes.io/docs/concepts/services-networking/service/#publishing-services-service-types
    # Possible enum values:
    #   - `"ClusterIP"` means a service will only be accessible inside the cluster, via the cluster IP.
    #   - `"ExternalName"` means a service consists of only a reference to an external name
    #       that kubedns or equivalent will return as a CNAME record, with no exposing or proxying of any pods involved.
    #   - `"LoadBalancer"` means a service will be exposed via an external load balancer
    #   - `"NodePort"` means a service will be exposed on one port of every node, in addition to 'ClusterIP' type.
    type: Optional[ServiceType] = None
    ## type == ClusterIP
    # clusterIP is the IP address of the service and is usually assigned randomly.
    # If an address is specified manually, is in-range (as per system configuration), and is not in use,
    # it will be allocated to the service; otherwise creation of the service will fail
    cluster_ip: Optional[str] = Field(None, alias="clusterIP")
    # ClusterIPs is a list of IP addresses assigned to this service, and are usually assigned randomly
    cluster_ips: Optional[List[str]] = Field(None, alias="clusterIPs")
    ## type == ExternalName
    # externalIPs is a list of IP addresses for which nodes in the cluster will also accept traffic for this service.
    # These IPs are not managed by Kubernetes. The user is responsible for ensuring that traffic arrives at a node
    # with this IP. An example is external load-balancers that are not part of the Kubernetes system.
    external_ips: Optional[List[str]] = Field(None, alias="externalIPs")
    # externalName is the external reference that discovery mechanisms will return as an alias for this
    # service (e.g. a DNS CNAME record). No proxying will be involved.
    # Must be a lowercase RFC-1123 hostname (https://tools.ietf.org/html/rfc1123) and requires
    # `type` to be "ExternalName".
    external_name: Optional[str] = Field(None, alias="externalName")
    # externalTrafficPolicy denotes if this Service desires to route external traffic
    # to node-local or cluster-wide endpoints.
    # "Local" preserves the client source IP and avoids a second hop for LoadBalancer and Nodeport type services,
    #   but risks potentially imbalanced traffic spreading.
    # "Cluster" obscures the client source IP and may cause a second hop to another node,
    # but should have good overall load-spreading.
    # Possible enum values:
    #   - `"Cluster"` specifies node-global (legacy) behavior.
    #   - `"Local"` specifies node-local endpoints behavior.
    external_traffic_policy: Optional[str] = Field(None, alias="externalTrafficPolicy")
    ## type == LoadBalancer
    # healthCheckNodePort specifies the healthcheck nodePort for the service.
    # This only applies when type is set to LoadBalancer and externalTrafficPolicy is set to Local.
    health_check_node_port: Optional[int] = Field(None, alias="healthCheckNodePort")
    # InternalTrafficPolicy specifies if the cluster internal traffic
    # should be routed to all endpoints or node-local endpoints only.
    # "Cluster" routes internal traffic to a Service to all endpoints.
    # "Local" routes traffic to node-local endpoints only, traffic is dropped if no node-local endpoints are ready.
    # The default value is "Cluster".
    internal_traffic_policy: Optional[str] = Field(None, alias="internalTrafficPolicy")
    # loadBalancerClass is the class of the load balancer implementation this Service belongs to.
    # If specified, the value of this field must be a label-style identifier, with an optional prefix,
    # e.g. "internal-vip" or "example.com/internal-vip". Unprefixed names are reserved for end-users.
    # This field can only be set when the Service type is 'LoadBalancer'.
    # If not set, the default load balancer implementation is used
    load_balancer_class: Optional[str] = Field(None, alias="loadBalancerClass")
    # Only applies to Service Type: LoadBalancer
    # LoadBalancer will get created with the IP specified in this field. This feature depends on
    # whether the underlying cloud-provider supports specifying the loadBalancerIP when a load balancer is created.
    # This field will be ignored if the cloud-provider does not support the feature.
    load_balancer_ip: Optional[str] = Field(None, alias="loadBalancerIP")
    # If specified and supported by the platform, this will restrict traffic through the cloud-provider load-balancer
    # will be restricted to the specified client IPs. This field will be ignored if the cloud-provider does not support.
    # More info: https://kubernetes.io/docs/tasks/access-application-cluster/create-external-load-balancer/
    load_balancer_source_ranges: Optional[List[str]] = Field(None, alias="loadBalancerSourceRanges")
    # allocateLoadBalancerNodePorts defines if NodePorts will be automatically allocated for services
    # with type LoadBalancer. Default is "true". It may be set to "false" if the cluster load-balancer
    # does not rely on NodePorts.
    allocate_load_balancer_node_ports: Optional[bool] = Field(None, alias="allocateLoadBalancerNodePorts")

    # The list of ports that are exposed by this service.
    # More info: https://kubernetes.io/docs/concepts/services-networking/service/#virtual-ips-and-service-proxies
    ports: List[ServicePort]
    publish_not_ready_addresses: Optional[bool] = Field(None, alias="publishNotReadyAddresses")
    # Route service traffic to pods with label keys and values matching this selector.
    # If empty or not present, the service is assumed to have an external process managing its endpoints,
    # which Kubernetes will not modify. Only applies to types ClusterIP, NodePort, and LoadBalancer.
    # Ignored if type is ExternalName. More info: https://kubernetes.io/docs/concepts/services-networking/service/
    selector: Dict[str, str]
    # Supports "ClientIP" and "None". Used to maintain session affinity.
    # Enable client IP based session affinity. Must be ClientIP or None. Defaults to None.
    session_affinity: Optional[str] = Field(None, alias="sessionAffinity")
    # sessionAffinityConfig contains the configurations of session affinity.
    # session_affinity_config: Optional[SessionAffinityConfig] = Field(None, alias="sessionAffinityConfig")

    @field_serializer("type")
    def get_type_value(self, v) -> Optional[str]:
        return v.value if v else None

    def get_k8s_object(self) -> V1ServiceSpec:
        # Return a V1ServiceSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service_spec.py
        _ports: Optional[List[V1ServicePort]] = None
        if self.ports:
            _ports = []
            for _port in self.ports:
                _ports.append(_port.get_k8s_object())

        _v1_service_spec = V1ServiceSpec(
            type=self.type.value if self.type else None,
            allocate_load_balancer_node_ports=self.allocate_load_balancer_node_ports,
            cluster_ip=self.cluster_ip,
            cluster_i_ps=self.cluster_ips,
            external_i_ps=self.external_ips,
            external_name=self.external_name,
            external_traffic_policy=self.external_traffic_policy,
            health_check_node_port=self.health_check_node_port,
            internal_traffic_policy=self.internal_traffic_policy,
            load_balancer_class=self.load_balancer_class,
            load_balancer_ip=self.load_balancer_ip,
            load_balancer_source_ranges=self.load_balancer_source_ranges,
            ports=_ports,
            publish_not_ready_addresses=self.publish_not_ready_addresses,
            selector=self.selector,
            session_affinity=self.session_affinity,
            # ip_families=self.ip_families,
            # ip_family_policy=self.ip_family_policy,
            # session_affinity_config=self.session_affinity_config,
        )
        return _v1_service_spec


class Service(K8sResource):
    """A service resource exposes an application running on a set of Pods
    as a network service.

    References:
    - Docs:
        https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#service-v1-core
        https://kubernetes.io/docs/concepts/services-networking/service/
    - Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service.py
    Notes:
        * The name of a Service object must be a valid DNS label name.
    """

    resource_type: str = "Service"

    spec: ServiceSpec

    # Only used to print the LoadBalancer DNS
    protocol: Optional[Literal["http", "https"]] = None

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> V1Service:
        """Creates a body for this Service"""

        # Return a V1Service object to create a ClusterRole
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_service.py
        _v1_service = V1Service(
            api_version=self.api_version.value,
            kind=self.kind.value,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_service

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1Service]]:
        """Reads Services from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        svc_list: Optional[V1ServiceList] = None
        if namespace:
            # logger.debug(f"Getting services for ns: {namespace}")
            svc_list = core_v1_api.list_namespaced_service(namespace=namespace)
        else:
            # logger.debug("Getting services for all namespaces")
            svc_list = core_v1_api.list_service_for_all_namespaces()

        services: Optional[List[V1Service]] = None
        if svc_list:
            services = svc_list.items
        # logger.debug(f"services: {services}")
        # logger.debug(f"services type: {type(services)}")
        return services

    def _create(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        k8s_object: V1Service = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_service: V1Service = core_v1_api.create_namespaced_service(
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Created: {}".format(v1_service))
        if v1_service.metadata.creation_timestamp is not None:
            logger.debug("Service Created")
            self.active_resource = v1_service
            return True
        logger.error("Service could not be created")
        return False

    def post_create(self, k8s_client: K8sApiClient) -> bool:
        from time import sleep

        if self.spec.type == ServiceType.LOAD_BALANCER:
            logger.info("Waiting for LoadBalancer DNS to be available")
            attempts = 0
            lb_dns = None
            while attempts < 10:
                attempts += 1
                svc: Optional[V1Service] = self._read(k8s_client=k8s_client)
                try:
                    if svc is not None:
                        if svc.status is not None:
                            if svc.status.load_balancer is not None:
                                if svc.status.load_balancer.ingress is not None:
                                    if svc.status.load_balancer.ingress[0] is not None:
                                        lb_dns = svc.status.load_balancer.ingress[0].hostname
                                        break
                    sleep(1)
                except AttributeError:
                    pass
            if lb_dns is None:
                logger.info("LoadBalancer DNS could not be found, please check the AWS console")
                return False
            else:
                if self.protocol is not None:
                    lb_dns = f"{self.protocol}://{lb_dns}"
                logger.info(f"LoadBalancer DNS: {lb_dns}")
        return True

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1Service]:
        """Returns the "Active" Service from the cluster"""

        namespace = self.get_namespace()
        active_resource: Optional[V1Service] = None
        active_resources: Optional[List[V1Service]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"Active Resources: {active_resources}")
        if active_resources is None:
            return None

        active_resources_dict = {_service.metadata.name: _service for _service in active_resources}

        svc_name = self.get_resource_name()
        if svc_name in active_resources_dict:
            active_resource = active_resources_dict[svc_name]
            self.active_resource = active_resource
            logger.debug(f"Found active {svc_name}")
        return active_resource

    def _update(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        svc_name = self.get_resource_name()
        k8s_object: V1Service = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Updating: {}".format(svc_name))
        v1_service: V1Service = core_v1_api.patch_namespaced_service(
            name=svc_name,
            namespace=namespace,
            body=k8s_object,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_service.to_dict(), indent=2)))
        if v1_service.metadata.creation_timestamp is not None:
            logger.debug("Service Updated")
            self.active_resource = v1_service
            return True
        logger.error("Service could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        svc_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(svc_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        delete_status: V1Status = core_v1_api.delete_namespaced_service(
            name=svc_name,
            namespace=namespace,
            async_req=self.async_req,
            pretty=self.pretty,
        )
        delete_service_status = delete_status.status
        logger.debug(f"Delete Status: {delete_service_status}")
        if isinstance(delete_service_status, V1ServiceStatus):
            if delete_service_status.conditions is None:
                logger.debug("Service Deleted")
                return True
        logger.error("Service could not be deleted")
        return False



from typing import Optional, Any, List

try:
    from sqlalchemy.dialects import postgresql
    from sqlalchemy.engine import create_engine, Engine
    from sqlalchemy.engine.row import Row
    from sqlalchemy.inspection import inspect
    from sqlalchemy.orm import Session, sessionmaker
    from sqlalchemy.schema import MetaData, Table, Column
    from sqlalchemy.sql.expression import text, select
    from sqlalchemy.types import DateTime, String
except ImportError:
    raise ImportError("`sqlalchemy` not installed")

try:
    import psycopg
except ImportError:
    raise ImportError("`psycopg` not installed")

from phi.conversation.schemas import ConversationRow
from phi.conversation.storage.base import ConversationStorage
from phi.utils.log import logger


class PgConversationStorage(ConversationStorage):
    def __init__(
        self,
        table_name: str,
        schema: Optional[str] = "llm",
        db_url: Optional[str] = None,
        db_engine: Optional[Engine] = None,
    ):
        """
        This class provides conversation storage using a postgres database.

        The following order is used to determine the database connection:
            1. Use the db_engine if provided
            2. Use the db_url

        :param table_name: The name of the table to store conversations in.
        :param schema: The name of the schema to store the table in.
        :param db_url: The database URL to connect to.
        :param db_engine: The database engine to use.
        """
        _engine: Optional[Engine] = db_engine
        if _engine is None and db_url is not None:
            _engine = create_engine(db_url)

        if _engine is None:
            raise ValueError("Must provide either db_url or db_engine")

        # Database attributes
        self.table_name: str = table_name
        self.schema: Optional[str] = schema
        self.db_url: Optional[str] = db_url
        self.db_engine: Engine = _engine
        self.metadata: MetaData = MetaData(schema=self.schema)

        # Database session
        self.Session: sessionmaker[Session] = sessionmaker(bind=self.db_engine)

        # Database table for storage
        self.table: Table = self.get_table()

    def get_table(self) -> Table:
        return Table(
            self.table_name,
            self.metadata,
            # Primary key for this conversation.
            Column("id", String, primary_key=True),
            # Conversation name
            Column("name", String),
            # Name and type of user participating in this conversation.
            Column("user_name", String),
            Column("user_type", String),
            # True if this conversation is active.
            Column("is_active", postgresql.BOOLEAN, server_default=text("true")),
            # -*- LLM data (name, model, etc.)
            Column("llm", postgresql.JSONB),
            # -*- Conversation memory
            Column("memory", postgresql.JSONB),
            # Metadata associated with this conversation.
            Column("meta_data", postgresql.JSONB),
            # Extra data associated with this conversation.
            Column("extra_data", postgresql.JSONB),
            # The timestamp of when this conversation was created.
            Column("created_at", DateTime(timezone=True), server_default=text("now()")),
            # The timestamp of when this conversation was last updated.
            Column("updated_at", DateTime(timezone=True), onupdate=text("now()")),
            extend_existing=True,
        )

    def table_exists(self) -> bool:
        logger.debug(f"Checking if table exists: {self.table.name}")
        try:
            return inspect(self.db_engine).has_table(self.table.name, schema=self.schema)
        except Exception as e:
            logger.error(e)
            return False

    def create(self) -> None:
        if not self.table_exists():
            if self.schema is not None:
                with self.Session() as sess, sess.begin():
                    logger.debug(f"Creating schema: {self.schema}")
                    sess.execute(text(f"create schema if not exists {self.schema};"))
            logger.debug(f"Creating table: {self.table_name}")
            self.table.create(self.db_engine)

    def _read(self, session: Session, conversation_id: str) -> Optional[Row[Any]]:
        stmt = select(self.table).where(self.table.c.id == conversation_id)
        try:
            return session.execute(stmt).first()
        except Exception as e:
            # Create table if it does not exist
            if e.__class__ == psycopg.errors.UndefinedTable:
                self.create()
        return None

    def read(self, conversation_id: str) -> Optional[ConversationRow]:
        with self.Session() as sess, sess.begin():
            existing_row: Optional[Row[Any]] = self._read(session=sess, conversation_id=conversation_id)
            return ConversationRow.model_validate(existing_row) if existing_row is not None else None

    def get_all_conversation_ids(self, user_name: str) -> List[str]:
        conversation_ids: List[str] = []
        try:
            with self.Session() as sess, sess.begin():
                # get all conversation ids for this user
                stmt = select(self.table).where(self.table.c.user_name == user_name)
                # order by created_at desc
                stmt = stmt.order_by(self.table.c.created_at.desc())
                # execute query
                rows = sess.execute(stmt).fetchall()
                for row in rows:
                    if row is not None and row.id is not None:
                        conversation_ids.append(row.id)
        except psycopg.errors.UndefinedTable:
            logger.debug(f"Table does not exist: {self.table.name}")
            pass
        return conversation_ids

    def get_all_conversations(self, user_name: str) -> List[ConversationRow]:
        conversations: List[ConversationRow] = []
        try:
            with self.Session() as sess, sess.begin():
                # get all conversation ids for this user
                stmt = select(self.table).where(self.table.c.user_name == user_name)
                # order by created_at desc
                stmt = stmt.order_by(self.table.c.created_at.desc())
                # execute query
                rows = sess.execute(stmt).fetchall()
                for row in rows:
                    if row.id is not None:
                        conversations.append(ConversationRow.model_validate(row))
        except psycopg.errors.UndefinedTable:
            logger.debug(f"Table does not exist: {self.table.name}")
            pass
        return conversations

    def upsert(self, conversation: ConversationRow) -> Optional[ConversationRow]:
        """
        Create a new conversation if it does not exist, otherwise update the existing conversation.
        """

        with self.Session() as sess, sess.begin():
            # Create an insert statement
            stmt = postgresql.insert(self.table).values(
                id=conversation.id,
                name=conversation.name,
                user_name=conversation.user_name,
                user_type=conversation.user_type,
                is_active=conversation.is_active,
                llm=conversation.llm,
                memory=conversation.memory,
                meta_data=conversation.meta_data,
                extra_data=conversation.extra_data,
            )

            # Define the upsert if the id already exists
            # See: https://docs.sqlalchemy.org/en/20/dialects/postgresql.html#postgresql-insert-on-conflict
            stmt = stmt.on_conflict_do_update(
                index_elements=["id"],
                set_=dict(
                    name=conversation.name,
                    user_name=conversation.user_name,
                    user_type=conversation.user_type,
                    is_active=conversation.is_active,
                    llm=conversation.llm,
                    memory=conversation.memory,
                    meta_data=conversation.meta_data,
                    extra_data=conversation.extra_data,
                ),  # The updated value for each column
            )

            try:
                sess.execute(stmt)
            except psycopg.errors.UndefinedTable:
                # Create table if it does not exist
                self.create()
                sess.execute(stmt)
        return self.read(conversation_id=conversation.id)

    def end(self, conversation_id: str) -> None:
        with self.Session() as sess, sess.begin():
            # Check if conversation exists in the database
            existing_row: Optional[Row[Any]] = self._read(session=sess, conversation_id=conversation_id)
            # If conversation exists, set is_active to False
            if existing_row is not None:
                stmt = self.table.update().where(self.table.c.id == conversation_id).values(is_active=False)
                sess.execute(stmt)

    def delete(self) -> None:
        if self.table_exists():
            logger.debug(f"Deleting table: {self.table_name}")
            self.table.drop(self.db_engine)



from datetime import datetime, timezone


def current_datetime() -> datetime:
    return datetime.now()


def current_datetime_utc() -> datetime:
    return datetime.now(timezone.utc)


def current_datetime_utc_str() -> str:
    return current_datetime_utc().strftime("%Y-%m-%dT%H:%M:%S")



from phi.k8s.app.airflow.base import (
    AirflowBase,
    AppVolumeType,
    ContainerContext,
    ServiceType,
    RestartPolicy,
    ImagePullPolicy,
)
from phi.k8s.app.airflow.webserver import AirflowWebserver
from phi.k8s.app.airflow.scheduler import AirflowScheduler
from phi.k8s.app.airflow.worker import AirflowWorker
from phi.k8s.app.airflow.flower import AirflowFlower



from typing import List, Optional, Union, Tuple

from phi.app.group import AppGroup
from phi.resource.group import ResourceGroup
from phi.aws.app.base import AwsApp
from phi.aws.app.context import AwsBuildContext
from phi.aws.api_client import AwsApiClient
from phi.aws.resource.base import AwsResource
from phi.infra.resources import InfraResources
from phi.utils.log import logger


class AwsResources(InfraResources):
    apps: Optional[List[Union[AwsApp, AppGroup]]] = None
    resources: Optional[List[Union[AwsResource, ResourceGroup]]] = None

    # -*- Cached Data
    _aws_region: Optional[str] = None
    _aws_profile: Optional[str] = None
    _api_client: Optional[AwsApiClient] = None

    @property
    def aws_region(self) -> Optional[str]:
        # Priority 1: Use aws_region from ResourceGroup (or cached value)
        if self._aws_region:
            return self._aws_region

        # Priority 2: Get aws_region from workspace settings
        if self.workspace_settings is not None and self.workspace_settings.aws_region is not None:
            self._aws_region = self.workspace_settings.aws_region
            return self._aws_region

        # Priority 3: Get aws_region from env
        from os import getenv
        from phi.constants import AWS_REGION_ENV_VAR

        aws_region_env = getenv(AWS_REGION_ENV_VAR)
        if aws_region_env is not None:
            logger.debug(f"{AWS_REGION_ENV_VAR}: {aws_region_env}")
            self._aws_region = aws_region_env
        return self._aws_region

    @aws_region.setter
    def aws_region(self, value: str) -> None:
        self._aws_region = value

    @property
    def aws_profile(self) -> Optional[str]:
        # Priority 1: Use aws_region from ResourceGroup (or cached value)
        if self._aws_profile:
            return self._aws_profile

        # Priority 2: Get aws_profile from workspace settings
        if self.workspace_settings is not None and self.workspace_settings.aws_profile is not None:
            self._aws_profile = self.workspace_settings.aws_profile
            return self._aws_profile

        # Priority 3: Get aws_profile from env
        from os import getenv
        from phi.constants import AWS_PROFILE_ENV_VAR

        aws_profile_env = getenv(AWS_PROFILE_ENV_VAR)
        if aws_profile_env is not None:
            logger.debug(f"{AWS_PROFILE_ENV_VAR}: {aws_profile_env}")
            self._aws_profile = aws_profile_env
        return self._aws_profile

    @aws_profile.setter
    def aws_profile(self, value: str) -> None:
        self._aws_profile = value

    @property
    def aws_client(self) -> AwsApiClient:
        if self._api_client is None:
            self._api_client = AwsApiClient(aws_region=self.aws_region, aws_profile=self.aws_profile)
        return self._api_client

    def create_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.aws.resource.types import AwsResourceInstallOrder

        logger.debug("-*- Creating AwsResources")
        # Build a list of AwsResources to create
        resources_to_create: List[AwsResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, AwsResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_create(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_create.append(resource_from_resource_group)
                elif isinstance(r, AwsResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_create(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        r.set_workspace_settings(workspace_settings=self.workspace_settings)
                        resources_to_create.append(r)

        # Build a list of AwsApps to create
        apps_to_create: List[AwsApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, AwsApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_create(group_filter=group_filter):
                                    apps_to_create.append(app_from_app_group)
                elif isinstance(app, AwsApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_create(group_filter=group_filter):
                        apps_to_create.append(app)

        # Get the list of AwsResources from the AwsApps
        if len(apps_to_create) > 0:
            logger.debug(f"Found {len(apps_to_create)} apps to create")
            for app in apps_to_create:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=AwsBuildContext(aws_region=self.aws_region, aws_profile=self.aws_profile)
                )
                if len(app_resources) > 0:
                    # If the app has dependencies, add the resources from the
                    # dependencies first to the list of resources to create
                    if app.depends_on is not None:
                        for dep in app.depends_on:
                            if isinstance(dep, AwsApp):
                                dep.set_workspace_settings(workspace_settings=self.workspace_settings)
                                dep_resources = dep.get_resources(
                                    build_context=AwsBuildContext(
                                        aws_region=self.aws_region, aws_profile=self.aws_profile
                                    )
                                )
                                if len(dep_resources) > 0:
                                    for dep_resource in dep_resources:
                                        if isinstance(dep_resource, AwsResource):
                                            resources_to_create.append(dep_resource)
                    # Add the resources from the app to the list of resources to create
                    for app_resource in app_resources:
                        if isinstance(app_resource, AwsResource) and app_resource.should_create(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_create.append(app_resource)

        # Sort the AwsResources in install order
        resources_to_create.sort(key=lambda x: AwsResourceInstallOrder.get(x.__class__.__name__, 5000))

        # Deduplicate AwsResources
        deduped_resources_to_create: List[AwsResource] = []
        for r in resources_to_create:
            if r not in deduped_resources_to_create:
                deduped_resources_to_create.append(r)

        # Implement dependency sorting
        final_aws_resources: List[AwsResource] = []
        logger.debug("-*- Building AwsResources dependency graph")
        for aws_resource in deduped_resources_to_create:
            # Logic to follow if resource has dependencies
            if aws_resource.depends_on is not None and len(aws_resource.depends_on) > 0:
                # Add the dependencies before the resource itself
                for dep in aws_resource.depends_on:
                    if isinstance(dep, AwsResource):
                        if dep not in final_aws_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {aws_resource.name}")
                            final_aws_resources.append(dep)

                # Add the resource to be created after its dependencies
                if aws_resource not in final_aws_resources:
                    logger.debug(f"-*- Adding {aws_resource.name}")
                    final_aws_resources.append(aws_resource)
            else:
                # Add the resource to be created if it has no dependencies
                if aws_resource not in final_aws_resources:
                    logger.debug(f"-*- Adding {aws_resource.name}")
                    final_aws_resources.append(aws_resource)

        # Track the total number of AwsResources to create for validation
        num_resources_to_create: int = len(final_aws_resources)
        num_resources_created: int = 0
        if num_resources_to_create == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- AWS resources to create:")
            for resource in final_aws_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            if self.aws_region:
                print_info(f"Region: {self.aws_region}")
            if self.aws_profile:
                print_info(f"Profile: {self.aws_profile}")
            print_info(f"Total {num_resources_to_create} resources")
            return 0, 0

        # Validate resources to be created
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to create:")
            for resource in final_aws_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            if self.aws_region:
                print_info(f"Region: {self.aws_region}")
            if self.aws_profile:
                print_info(f"Profile: {self.aws_profile}")
            print_info(f"Total {num_resources_to_create} resources")
            confirm = confirm_yes_no("\nConfirm deploy")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping create")
                print_info("-*-")
                return 0, 0

        for resource in final_aws_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            # logger.debug(resource)
            try:
                _resource_created = resource.create(aws_client=self.aws_client)
                if _resource_created:
                    num_resources_created += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_create_failure:
                        return num_resources_created, num_resources_to_create
            except Exception as e:
                logger.error(f"Failed to create {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources created: {num_resources_created}/{num_resources_to_create}")
        if num_resources_to_create != num_resources_created:
            logger.error(
                f"Resources created: {num_resources_created} do not match resources required: {num_resources_to_create}"
            )  # noqa: E501
        return num_resources_created, num_resources_to_create

    def delete_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.aws.resource.types import AwsResourceInstallOrder

        logger.debug("-*- Deleting AwsResources")

        # Build a list of AwsResources to delete
        resources_to_delete: List[AwsResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, AwsResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_delete(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_delete.append(resource_from_resource_group)
                elif isinstance(r, AwsResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_delete(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        r.set_workspace_settings(workspace_settings=self.workspace_settings)
                        resources_to_delete.append(r)

        # Build a list of AwsApps to delete
        apps_to_delete: List[AwsApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, AwsApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_delete(group_filter=group_filter):
                                    apps_to_delete.append(app_from_app_group)
                elif isinstance(app, AwsApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_delete(group_filter=group_filter):
                        apps_to_delete.append(app)

        # Get the list of AwsResources from the AwsApps
        if len(apps_to_delete) > 0:
            logger.debug(f"Found {len(apps_to_delete)} apps to delete")
            for app in apps_to_delete:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=AwsBuildContext(aws_region=self.aws_region, aws_profile=self.aws_profile)
                )
                if len(app_resources) > 0:
                    for app_resource in app_resources:
                        if isinstance(app_resource, AwsResource) and app_resource.should_delete(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_delete.append(app_resource)

        # Sort the AwsResources in install order
        resources_to_delete.sort(key=lambda x: AwsResourceInstallOrder.get(x.__class__.__name__, 5000), reverse=True)

        # Deduplicate AwsResources
        deduped_resources_to_delete: List[AwsResource] = []
        for r in resources_to_delete:
            if r not in deduped_resources_to_delete:
                deduped_resources_to_delete.append(r)

        # Implement dependency sorting
        final_aws_resources: List[AwsResource] = []
        logger.debug("-*- Building AwsResources dependency graph")
        for aws_resource in deduped_resources_to_delete:
            # Logic to follow if resource has dependencies
            if aws_resource.depends_on is not None and len(aws_resource.depends_on) > 0:
                # 1. Reverse the order of dependencies
                aws_resource.depends_on.reverse()

                # 2. Remove the dependencies if they are already added to the final_aws_resources
                for dep in aws_resource.depends_on:
                    if dep in final_aws_resources:
                        logger.debug(f"-*- Removing {dep.name}, dependency of {aws_resource.name}")
                        final_aws_resources.remove(dep)

                # 3. Add the resource to be deleted before its dependencies
                if aws_resource not in final_aws_resources:
                    logger.debug(f"-*- Adding {aws_resource.name}")
                    final_aws_resources.append(aws_resource)

                # 4. Add the dependencies back in reverse order
                for dep in aws_resource.depends_on:
                    if isinstance(dep, AwsResource):
                        if dep not in final_aws_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {aws_resource.name}")
                            final_aws_resources.append(dep)
            else:
                # Add the resource to be deleted if it has no dependencies
                if aws_resource not in final_aws_resources:
                    logger.debug(f"-*- Adding {aws_resource.name}")
                    final_aws_resources.append(aws_resource)

        # Track the total number of AwsResources to delete for validation
        num_resources_to_delete: int = len(final_aws_resources)
        num_resources_deleted: int = 0
        if num_resources_to_delete == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- AWS resources to delete:")
            for resource in final_aws_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            if self.aws_region:
                print_info(f"Region: {self.aws_region}")
            if self.aws_profile:
                print_info(f"Profile: {self.aws_profile}")
            print_info(f"Total {num_resources_to_delete} resources")
            return 0, 0

        # Validate resources to be deleted
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to delete:")
            for resource in final_aws_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            if self.aws_region:
                print_info(f"Region: {self.aws_region}")
            if self.aws_profile:
                print_info(f"Profile: {self.aws_profile}")
            print_info(f"Total {num_resources_to_delete} resources")
            confirm = confirm_yes_no("\nConfirm delete")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping delete")
                print_info("-*-")
                return 0, 0

        for resource in final_aws_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            # logger.debug(resource)
            try:
                _resource_deleted = resource.delete(aws_client=self.aws_client)
                if _resource_deleted:
                    num_resources_deleted += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_delete_failure:
                        return num_resources_deleted, num_resources_to_delete
            except Exception as e:
                logger.error(f"Failed to delete {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources deleted: {num_resources_deleted}/{num_resources_to_delete}")
        if num_resources_to_delete != num_resources_deleted:
            logger.error(
                f"Resources deleted: {num_resources_deleted} do not match resources required: {num_resources_to_delete}"
            )  # noqa: E501
        return num_resources_deleted, num_resources_to_delete

    def update_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
        dry_run: Optional[bool] = False,
        auto_confirm: Optional[bool] = False,
        force: Optional[bool] = None,
    ) -> Tuple[int, int]:
        from phi.cli.console import print_info, print_heading, confirm_yes_no
        from phi.aws.resource.types import AwsResourceInstallOrder

        logger.debug("-*- Updating AwsResources")

        # Build a list of AwsResources to update
        resources_to_update: List[AwsResource] = []
        if self.resources is not None:
            for r in self.resources:
                if isinstance(r, ResourceGroup):
                    resources_from_resource_group = r.get_resources()
                    if len(resources_from_resource_group) > 0:
                        for resource_from_resource_group in resources_from_resource_group:
                            if isinstance(resource_from_resource_group, AwsResource):
                                resource_from_resource_group.set_workspace_settings(
                                    workspace_settings=self.workspace_settings
                                )
                                if resource_from_resource_group.group is None and self.name is not None:
                                    resource_from_resource_group.group = self.name
                                if resource_from_resource_group.should_update(
                                    group_filter=group_filter,
                                    name_filter=name_filter,
                                    type_filter=type_filter,
                                ):
                                    resources_to_update.append(resource_from_resource_group)
                elif isinstance(r, AwsResource):
                    r.set_workspace_settings(workspace_settings=self.workspace_settings)
                    if r.group is None and self.name is not None:
                        r.group = self.name
                    if r.should_update(
                        group_filter=group_filter,
                        name_filter=name_filter,
                        type_filter=type_filter,
                    ):
                        r.set_workspace_settings(workspace_settings=self.workspace_settings)
                        resources_to_update.append(r)

        # Build a list of AwsApps to update
        apps_to_update: List[AwsApp] = []
        if self.apps is not None:
            for app in self.apps:
                if isinstance(app, AppGroup):
                    apps_from_app_group = app.get_apps()
                    if len(apps_from_app_group) > 0:
                        for app_from_app_group in apps_from_app_group:
                            if isinstance(app_from_app_group, AwsApp):
                                if app_from_app_group.group is None and self.name is not None:
                                    app_from_app_group.group = self.name
                                if app_from_app_group.should_update(group_filter=group_filter):
                                    apps_to_update.append(app_from_app_group)
                elif isinstance(app, AwsApp):
                    if app.group is None and self.name is not None:
                        app.group = self.name
                    if app.should_update(group_filter=group_filter):
                        apps_to_update.append(app)

        # Get the list of AwsResources from the AwsApps
        if len(apps_to_update) > 0:
            logger.debug(f"Found {len(apps_to_update)} apps to update")
            for app in apps_to_update:
                app.set_workspace_settings(workspace_settings=self.workspace_settings)
                app_resources = app.get_resources(
                    build_context=AwsBuildContext(aws_region=self.aws_region, aws_profile=self.aws_profile)
                )
                if len(app_resources) > 0:
                    for app_resource in app_resources:
                        if isinstance(app_resource, AwsResource) and app_resource.should_update(
                            group_filter=group_filter, name_filter=name_filter, type_filter=type_filter
                        ):
                            resources_to_update.append(app_resource)

        # Sort the AwsResources in install order
        resources_to_update.sort(key=lambda x: AwsResourceInstallOrder.get(x.__class__.__name__, 5000))

        # Deduplicate AwsResources
        deduped_resources_to_update: List[AwsResource] = []
        for r in resources_to_update:
            if r not in deduped_resources_to_update:
                deduped_resources_to_update.append(r)

        # Implement dependency sorting
        final_aws_resources: List[AwsResource] = []
        logger.debug("-*- Building AwsResources dependency graph")
        for aws_resource in deduped_resources_to_update:
            # Logic to follow if resource has dependencies
            if aws_resource.depends_on is not None and len(aws_resource.depends_on) > 0:
                # Add the dependencies before the resource itself
                for dep in aws_resource.depends_on:
                    if isinstance(dep, AwsResource):
                        if dep not in final_aws_resources:
                            logger.debug(f"-*- Adding {dep.name}, dependency of {aws_resource.name}")
                            final_aws_resources.append(dep)

                # Add the resource to be created after its dependencies
                if aws_resource not in final_aws_resources:
                    logger.debug(f"-*- Adding {aws_resource.name}")
                    final_aws_resources.append(aws_resource)
            else:
                # Add the resource to be created if it has no dependencies
                if aws_resource not in final_aws_resources:
                    logger.debug(f"-*- Adding {aws_resource.name}")
                    final_aws_resources.append(aws_resource)

        # Track the total number of AwsResources to update for validation
        num_resources_to_update: int = len(final_aws_resources)
        num_resources_updated: int = 0
        if num_resources_to_update == 0:
            return 0, 0

        if dry_run:
            print_heading("--**- AWS resources to update:")
            for resource in final_aws_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            if self.aws_region:
                print_info(f"Region: {self.aws_region}")
            if self.aws_profile:
                print_info(f"Profile: {self.aws_profile}")
            print_info(f"Total {num_resources_to_update} resources")
            return 0, 0

        # Validate resources to be updated
        if not auto_confirm:
            print_heading("\n--**-- Confirm resources to update:")
            for resource in final_aws_resources:
                print_info(f"  -+-> {resource.get_resource_type()}: {resource.get_resource_name()}")
            print_info("")
            if self.aws_region:
                print_info(f"Region: {self.aws_region}")
            if self.aws_profile:
                print_info(f"Profile: {self.aws_profile}")
            print_info(f"Total {num_resources_to_update} resources")
            confirm = confirm_yes_no("\nConfirm patch")
            if not confirm:
                print_info("-*-")
                print_info("-*- Skipping patch")
                print_info("-*-")
                return 0, 0

        for resource in final_aws_resources:
            print_info(f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}")
            if force is True:
                resource.force = True
            # logger.debug(resource)
            try:
                _resource_updated = resource.update(aws_client=self.aws_client)
                if _resource_updated:
                    num_resources_updated += 1
                else:
                    if self.workspace_settings is not None and not self.workspace_settings.continue_on_patch_failure:
                        return num_resources_updated, num_resources_to_update
            except Exception as e:
                logger.error(f"Failed to update {resource.get_resource_type()}: {resource.get_resource_name()}")
                logger.error(e)
                logger.error("Please fix and try again...")

        print_heading(f"\n--**-- Resources updated: {num_resources_updated}/{num_resources_to_update}")
        if num_resources_to_update != num_resources_updated:
            logger.error(
                f"Resources updated: {num_resources_updated} do not match resources required: {num_resources_to_update}"
            )  # noqa: E501
        return num_resources_updated, num_resources_to_update

    def save_resources(
        self,
        group_filter: Optional[str] = None,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Tuple[int, int]:
        raise NotImplementedError



from pathlib import Path
from typing import List

from phi.document.base import Document
from phi.document.reader.base import Reader
from phi.aws.resource.s3.object import S3Object
from phi.utils.log import logger


class S3TextReader(Reader):
    """Reader for text files on S3"""

    def read(self, s3_object: S3Object) -> List[Document]:
        if not s3_object:
            raise ValueError("No s3_object provided")

        try:
            import textract  # noqa: F401
        except ImportError:
            raise ImportError("`textract` not installed")

        try:
            logger.info(f"Reading: {s3_object.uri}")

            obj_name = s3_object.name.split("/")[-1]
            temporary_file = Path("storage").joinpath(obj_name)
            s3_object.download(temporary_file)

            logger.info(f"Parsing: {temporary_file}")
            doc_name = s3_object.name.split("/")[-1].split(".")[0].replace("/", "_").replace(" ", "_")
            doc_content = textract.process(temporary_file)
            documents = [
                Document(
                    name=doc_name,
                    content=doc_content.decode("utf-8"),
                )
            ]
            if self.chunk:
                chunked_documents = []
                for document in documents:
                    chunked_documents.extend(self.chunk_document(document))
                return chunked_documents

            logger.debug(f"Deleting: {temporary_file}")
            temporary_file.unlink()
            return documents
        except Exception as e:
            logger.error(f"Error reading: {s3_object.uri}: {e}")
        return []



from textwrap import dedent
from typing import Optional, Any, Dict, List
from typing_extensions import Literal

from phi.aws.api_client import AwsApiClient
from phi.aws.resource.base import AwsResource
from phi.aws.resource.ecs.container import EcsContainer
from phi.aws.resource.ecs.volume import EcsVolume
from phi.aws.resource.iam.role import IamRole
from phi.aws.resource.iam.policy import IamPolicy
from phi.cli.console import print_info
from phi.utils.log import logger


class EcsTaskDefinition(AwsResource):
    """
    Reference:
    - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/register_task_definition.html
    """

    resource_type: Optional[str] = "TaskDefinition"
    service_name: str = "ecs"

    # Name of the task definition.
    # Used as task definition family.
    name: str
    # The family for a task definition.
    # Use name as family if not provided
    # You can use it track multiple versions of the same task definition.
    # The family is used as a name for your task definition.
    family: Optional[str] = None
    # Networking mode to use for the containers in the task.
    # The valid values are none, bridge, awsvpc, and host.
    # If no network mode is specified, the default is bridge.
    network_mode: Optional[Literal["bridge", "host", "awsvpc", "none"]] = None
    # A list of container definitions that describe the different containers that make up the task.
    containers: Optional[List[EcsContainer]] = None
    volumes: Optional[List[EcsVolume]] = None
    placement_constraints: Optional[List[Dict[str, Any]]] = None
    requires_compatibilities: Optional[List[str]] = None
    cpu: Optional[str] = None
    memory: Optional[str] = None
    tags: Optional[List[Dict[str, str]]] = None
    pid_mode: Optional[Literal["host", "task"]] = None
    ipc_mode: Optional[Literal["host", "task", "none"]] = None
    proxy_configuration: Optional[Dict[str, Any]] = None
    inference_accelerators: Optional[List[Dict[str, Any]]] = None
    ephemeral_storage: Optional[Dict[str, Any]] = None
    runtime_platform: Optional[Dict[str, Any]] = None

    # Amazon ECS IAM roles
    # The short name or full Amazon Resource Name (ARN) of the IAM role that containers in this task can assume.
    # The permissions granted in this IAM role are assumed by the containers running in the task.
    # For permissions that Amazon ECS needs to pull container images, see execution_role_arn
    # If your containerized applications need to call AWS APIs, they must sign their
    # AWS API requests with AWS credentials, and a task IAM role provides a strategy for managing credentials
    # for your applications to use
    task_role_arn: Optional[str] = None
    # If task_role_arn is None, a default role is created if create_task_role is True
    create_task_role: bool = True
    # Name for the default role when task_role_arn is None, use "name-task-role" if not provided
    task_role_name: Optional[str] = None
    # Provide a list of policy ARNs to attach to the role
    add_policy_arns_to_task_role: Optional[List[str]] = None
    # Provide a list of IamPolicy to attach to the task role
    add_policies_to_task_role: Optional[List[IamPolicy]] = None
    # Add ecs_exec_policy to task role
    add_exec_access_to_task: bool = False
    # Add secret access to task role
    add_secret_access_to_task: bool = False
    # Add s3 access to task role
    add_s3_access_to_task: bool = False

    # The Amazon Resource Name (ARN) of the task execution role that grants the Amazon ECS container agent permission
    # to make Amazon Web Services API calls on your behalf. The task execution IAM role is required depending on the
    # requirements of your task.
    execution_role_arn: Optional[str] = None
    # If execution_role_arn is None, a default role is created if create_execution_role is True
    create_execution_role: bool = True
    # Name for the default role when execution_role_arn is None, use "name-execution-role" if not provided
    execution_role_name: Optional[str] = None
    # Provide a list of policy ARNs to attach to the role
    add_policy_arns_to_execution_role: Optional[List[str]] = None
    # Provide a list of IamPolicy to attach to the execution role
    add_policies_to_execution_role: Optional[List[IamPolicy]] = None
    # Add policy to read secrets to execution role
    add_secret_access_to_ecs: bool = False

    def get_task_family(self):
        return self.family or self.name

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Create EcsTaskDefinition"""
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Get task role arn
        task_role_arn = self.task_role_arn
        if task_role_arn is None and self.create_task_role:
            # Create the IamRole and get task_role_arn
            task_role = self.get_task_role()
            try:
                task_role.create(aws_client)
                task_role_arn = task_role.read(aws_client).arn
                print_info(f"ARN for {task_role.name}: {task_role_arn}")
            except Exception as e:
                logger.error("IamRole creation failed, please fix and try again")
                logger.error(e)
                return False

        # Step 2: Get execution role arn
        execution_role_arn = self.execution_role_arn
        if execution_role_arn is None and self.create_execution_role:
            # Create the IamRole and get execution_role_arn
            execution_role = self.get_execution_role()
            try:
                execution_role.create(aws_client)
                execution_role_arn = execution_role.read(aws_client).arn
                print_info(f"ARN for {execution_role.name}: {execution_role_arn}")
            except Exception as e:
                logger.error("IamRole creation failed, please fix and try again")
                logger.error(e)
                return False

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if task_role_arn is not None:
            not_null_args["taskRoleArn"] = task_role_arn
        if execution_role_arn is not None:
            not_null_args["executionRoleArn"] = execution_role_arn
        if self.network_mode is not None:
            not_null_args["networkMode"] = self.network_mode
        if self.containers is not None:
            container_definitions = [c.get_container_definition(aws_client=aws_client) for c in self.containers]
            not_null_args["containerDefinitions"] = container_definitions
        if self.volumes is not None:
            volume_definitions = [v.get_volume_definition() for v in self.volumes]
            not_null_args["volumes"] = volume_definitions
        if self.placement_constraints is not None:
            not_null_args["placementConstraints"] = self.placement_constraints
        if self.requires_compatibilities is not None:
            not_null_args["requiresCompatibilities"] = self.requires_compatibilities
        if self.cpu is not None:
            not_null_args["cpu"] = self.cpu
        if self.memory is not None:
            not_null_args["memory"] = self.memory
        if self.tags is not None:
            not_null_args["tags"] = self.tags
        if self.pid_mode is not None:
            not_null_args["pidMode"] = self.pid_mode
        if self.ipc_mode is not None:
            not_null_args["ipcMode"] = self.ipc_mode
        if self.proxy_configuration is not None:
            not_null_args["proxyConfiguration"] = self.proxy_configuration
        if self.inference_accelerators is not None:
            not_null_args["inferenceAccelerators"] = self.inference_accelerators
        if self.ephemeral_storage is not None:
            not_null_args["ephemeralStorage"] = self.ephemeral_storage
        if self.runtime_platform is not None:
            not_null_args["runtimePlatform"] = self.runtime_platform

        # Register EcsTaskDefinition
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.register_task_definition(
                family=self.get_task_family(),
                **not_null_args,
            )
            logger.debug(f"EcsTaskDefinition: {create_response}")
            resource_dict = create_response.get("taskDefinition", {})

            # Validate resource creation
            if resource_dict is not None:
                self.active_resource = create_response
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be created.")
            logger.error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Read EcsTaskDefinition"""
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_task_definition(taskDefinition=self.get_task_family())
            logger.debug(f"EcsTaskDefinition: {describe_response}")
            resource = describe_response.get("taskDefinition", None)
            if resource is not None:
                # compare the task definition with the current state
                # if there is a difference, create a new task definition
                # TODO: fix the task_definition_up_to_date function
                # if self.task_definition_up_to_date(task_definition=resource):
                self.active_resource = resource
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            logger.error(f"Error reading {self.get_resource_type()}.")
            logger.error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Delete EcsTaskDefinition"""
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Delete the task role
        if self.task_role_arn is None and self.create_task_role:
            task_role = self.get_task_role()
            try:
                task_role.delete(aws_client)
            except Exception as e:
                logger.error("IamRole deletion failed, please try again or delete manually")
                logger.error(e)

        # Step 2: Delete the execution role
        if self.execution_role_arn is None and self.create_execution_role:
            execution_role = self.get_execution_role()
            try:
                execution_role.delete(aws_client)
            except Exception as e:
                logger.error("IamRole deletion failed, please try again or delete manually")
                logger.error(e)

        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            # Get the task definition revisions
            list_response = service_client.list_task_definitions(familyPrefix=self.get_task_family(), sort="DESC")
            logger.debug(f"EcsTaskDefinition: {list_response}")
            task_definition_arns = list_response.get("taskDefinitionArns", [])
            if task_definition_arns:
                # Delete all revisions
                for task_definition_arn in task_definition_arns:
                    service_client.deregister_task_definition(taskDefinition=task_definition_arn)
                print_info(f"EcsTaskDefinition deleted: {self.get_resource_name()}")
                return True
        except Exception as e:
            logger.error(f"{self.get_resource_type()} could not be deleted.")
            logger.error("Please try again or delete resources manually.")
            logger.error(e)
        return False

    def _update(self, aws_client: AwsApiClient) -> bool:
        """Update EcsTaskDefinition"""
        print_info(f"Updating {self.get_resource_type()}: {self.get_resource_name()}")

        return self._create(aws_client)

    def get_task_role(self) -> IamRole:
        policy_arns = [
            "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy",
            "arn:aws:iam::aws:policy/CloudWatchFullAccess",
        ]
        if self.add_policy_arns_to_task_role is not None and isinstance(self.add_policy_arns_to_task_role, list):
            policy_arns.extend(self.add_policy_arns_to_task_role)

        policies = []
        if self.add_exec_access_to_task:
            ecs_exec_policy = IamPolicy(
                name=f"{self.name}-task-exec-policy",
                policy_document=dedent(
                    """\
                {
                   "Version": "2012-10-17",
                   "Statement": [
                       {
                       "Effect": "Allow",
                       "Action": [
                            "ssmmessages:CreateControlChannel",
                            "ssmmessages:CreateDataChannel",
                            "ssmmessages:OpenControlChannel",
                            "ssmmessages:OpenDataChannel"
                       ],
                      "Resource": "*"
                      }
                   ]
                }
                """
                ),
            )
            policies.append(ecs_exec_policy)
        if self.add_secret_access_to_task:
            policy_arns.append("arn:aws:iam::aws:policy/SecretsManagerReadWrite")
        if self.add_s3_access_to_task:
            policy_arns.append("arn:aws:iam::aws:policy/AmazonS3FullAccess")
        if self.add_policies_to_task_role:
            policies.extend(self.add_policies_to_task_role)

        return IamRole(
            name=self.task_role_name or f"{self.name}-task-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "ecs-tasks.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policies=policies,
            policy_arns=policy_arns,
        )

    def get_execution_role(self) -> IamRole:
        policy_arns = [
            "arn:aws:iam::aws:policy/service-role/AmazonECSTaskExecutionRolePolicy",
            "arn:aws:iam::aws:policy/CloudWatchFullAccess",
        ]
        if self.add_policy_arns_to_execution_role is not None and isinstance(
            self.add_policy_arns_to_execution_role, list
        ):
            policy_arns.extend(self.add_policy_arns_to_execution_role)

        policies = []
        if self.add_secret_access_to_ecs:
            ecs_secret_policy = IamPolicy(
                name=f"{self.name}-ecs-secret-policy",
                policy_document=dedent(
                    """\
                {
                   "Version": "2012-10-17",
                   "Statement": [
                       {
                       "Effect": "Allow",
                       "Action": [
                            "secretsmanager:GetSecretValue",
                            "secretsmanager:DescribeSecret",
                            "secretsmanager:ListSecretVersionIds"
                       ],
                      "Resource": "*"
                      }
                   ]
                }
                """
                ),
            )
            policies.append(ecs_secret_policy)
        if self.add_policies_to_execution_role:
            policies.extend(self.add_policies_to_execution_role)

        return IamRole(
            name=self.execution_role_name or f"{self.name}-execution-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "ecs-tasks.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policies=policies,
            policy_arns=policy_arns,
        )

    # def task_definition_up_to_date(self, task_definition: Dict[str, Any]) -> bool:
    #     """Return True if task_definition from the cluster matches the current state"""
    #
    #     # Validate container definitions
    #     if self.containers is not None:
    #         container_definitions_from_api = task_definition.get("containerDefinitions")
    #         # Compare the container definitions from the api with the current containers
    #         # The order of the container definitions should also match
    #         if container_definitions_from_api is not None and len(container_definitions_from_api) == len(
    #             self.containers
    #         ):
    #             for i, container in enumerate(self.containers):
    #                 if not container.container_definition_up_to_date(
    #                     container_definition=container_definitions_from_api[i]
    #                 ):
    #                     logger.debug("Container definitions not up to date")
    #                     return False
    #         else:
    #             logger.debug("Container definitions not up to date")
    #             return False
    #
    #     # Validate volumes
    #     if self.volumes is not None:
    #         volume_definitions_from_api = task_definition.get("volumes")
    #         # Compare the volume definitions from the api with the current volumes
    #         # The order of the volume definitions should also match
    #         if volume_definitions_from_api is not None and len(volume_definitions_from_api) == len(
    #             self.volumes
    #         ):
    #             for i, volume in enumerate(self.volumes):
    #                 if not volume.volume_definition_up_to_date(
    #                     volume_definition=volume_definitions_from_api[i]
    #                 ):
    #                     logger.debug("Volume definitions not up to date")
    #                     return False
    #         else:
    #             logger.debug("Volume definitions not up to date")
    #             return False
    #
    #     # Validate other properties
    #     if self.task_role_arn is not None:
    #         if self.task_role_arn != task_definition.get("taskRoleArn"):
    #             logger.debug("{} != {}".format(self.task_role_arn, task_definition.get("taskRoleArn")))
    #             return False
    #     if self.execution_role_arn is not None:
    #         if self.execution_role_arn != task_definition.get("executionRoleArn"):
    #             logger.debug(
    #                 "{} != {}".format(self.execution_role_arn, task_definition.get("executionRoleArn"))
    #             )
    #             return False
    #     if self.network_mode is not None:
    #         if self.network_mode != task_definition.get("networkMode"):
    #             logger.debug("{} != {}".format(self.network_mode, task_definition.get("networkMode")))
    #             return False
    #     if self.placement_constraints is not None:
    #         if self.placement_constraints != task_definition.get("placementConstraints"):
    #             logger.debug(
    #                 "{} != {}".format(
    #                     self.placement_constraints,
    #                     task_definition.get("placementConstraints"),
    #                 )
    #             )
    #             return False
    #     if self.requires_compatibilities is not None:
    #         if self.requires_compatibilities != task_definition.get("requiresCompatibilities"):
    #             logger.debug(
    #                 "{} != {}".format(
    #                     self.requires_compatibilities,
    #                     task_definition.get("requiresCompatibilities"),
    #                 )
    #             )
    #             return False
    #     if self.cpu is not None:
    #         if self.cpu != task_definition.get("cpu"):
    #             logger.debug("{} != {}".format(self.cpu, task_definition.get("cpu")))
    #             return False
    #     if self.memory is not None:
    #         if self.memory != task_definition.get("memory"):
    #             logger.debug("{} != {}".format(self.memory, task_definition.get("memory")))
    #             return False
    #     if self.tags is not None:
    #         if self.tags != task_definition.get("tags"):
    #             logger.debug("{} != {}".format(self.tags, task_definition.get("tags")))
    #             return False
    #     if self.pid_mode is not None:
    #         if self.pid_mode != task_definition.get("pidMode"):
    #             logger.debug("{} != {}".format(self.pid_mode, task_definition.get("pidMode")))
    #             return False
    #     if self.ipc_mode is not None:
    #         if self.ipc_mode != task_definition.get("ipcMode"):
    #             logger.debug("{} != {}".format(self.ipc_mode, task_definition.get("ipcMode")))
    #             return False
    #     if self.proxy_configuration is not None:
    #         if self.proxy_configuration != task_definition.get("proxyConfiguration"):
    #             logger.debug(
    #                 "{} != {}".format(
    #                     self.proxy_configuration,
    #                     task_definition.get("proxyConfiguration"),
    #                 )
    #             )
    #             return False
    #     if self.inference_accelerators is not None:
    #         if self.inference_accelerators != task_definition.get("inferenceAccelerators"):
    #             logger.debug(
    #                 "{} != {}".format(
    #                     self.inference_accelerators,
    #                     task_definition.get("inferenceAccelerators"),
    #                 )
    #             )
    #             return False
    #     if self.ephemeral_storage is not None:
    #         if self.ephemeral_storage != task_definition.get("ephemeralStorage"):
    #             logger.debug(
    #                 "{} != {}".format(self.ephemeral_storage, task_definition.get("ephemeralStorage"))
    #             )
    #             return False
    #     if self.runtime_platform is not None:
    #         if self.runtime_platform != task_definition.get("runtimePlatform"):
    #             logger.debug("{} != {}".format(self.runtime_platform, task_definition.get("runtimePlatform")))
    #             return False
    #
    #     return True



from pathlib import Path
from typing import List

from phi.document.base import Document
from phi.document.reader.base import Reader
from phi.utils.log import logger


class TextReader(Reader):
    """Reader for Text files"""

    def read(self, path: Path) -> List[Document]:
        if not path:
            raise ValueError("No path provided")

        if not path.exists():
            raise FileNotFoundError(f"Could not find file: {path}")

        try:
            import textract  # noqa: F401
        except ImportError:
            raise ImportError("`textract` not installed")

        try:
            logger.info(f"Reading: {path}")
            doc_name = path.name.split("/")[-1].split(".")[0].replace("/", "_").replace(" ", "_")
            doc_content = textract.process(path)
            documents = [
                Document(
                    name=doc_name,
                    content=doc_content.decode("utf-8"),
                )
            ]
            if self.chunk:
                chunked_documents = []
                for document in documents:
                    chunked_documents.extend(self.chunk_document(document))
                return chunked_documents
            return documents
        except Exception as e:
            logger.error(f"Error reading: {path}: {e}")
        return []
